import typing as ty
import json
from pydra.compose import python
from fileformats.generic import File
from fileformats.core import FileSet


def define_identity(field_names: ty.List[str]) -> python.Task:
    """Define a function task that returns the input fields as output.

    Parameters
    ----------
    fields : list of str
        List of field names to be used in the task.

    Returns
    -------
    FunctionTask
        A function task that takes the input fields and returns them as output.
    """

    @python.define(inputs=field_names, outputs=field_names)
    def Identity(**kwargs: ty.Any) -> ty.Tuple[ty.Any, ...]:
        """Identity function that returns the input fields as output."""
        return tuple(kwargs[n] for n in field_names)

    return Identity


@python.define(outputs=["out_file"])
def IdentityConverter(in_file: FileSet) -> FileSet:
    return in_file


@python.define
def ExtractFromJson(in_file: File, field_name: str) -> ty.Any:
    with open(in_file) as f:
        dct = json.load(f)
    return dct[field_name]  # FIXME: Should use JSONpath syntax



import os
import docker
from pathlib import Path
from copy import deepcopy
from pydra.utils import get_fields
from frametree.common import FileSystem, Samples
from pydra2app.core.image import App, P2AImage
from pydra2app.core import PACKAGE_NAME
from conftest import TestDatasetBlueprint


def test_native_python_install(tmp_path):

    SAMPLE_INDEX = "1"
    OUTPUT_COL_NAME = "printed_version"

    dataset_dir = tmp_path / "dataset"
    sample_dir = dataset_dir / SAMPLE_INDEX
    sample_dir.mkdir(parents=True)
    sample_file = sample_dir / "sample.txt"
    sample_file.write_text("sample")

    dataset = FileSystem().define_frameset(dataset_dir, axes=Samples)
    dataset.save()

    test_spec = {
        "name": "native_python_test",
        "title": "a test image spec",
        "commands": {
            "python-test-command": {
                "task": {
                    "type": "shell",
                    "executable": [
                        "pydra2app",
                        "--version<print_version>",
                    ],
                    "inputs": {
                        "dummy": {
                            "type": int | None,
                            "help": "not actually used",
                            "argstr": None,  # won't be printed to the command line
                        }
                    },
                },
                "row_frequency": "common:Samples[sample]",
            },
        },
        "version": "1.0",
        "packages": {
            "system": ["vim", "git"],  # just to test it out
            "pip": {
                "pydra2app": None,
                "frametree": None,
                "pydra": None,
            },  # just to test out the
        },
        "base_image": {
            "name": "python",
            "tag": "3.12.5-slim-bookworm",
            "python": "python3",
            "package_manager": "apt",
            "conda_env": None,
        },
        "authors": [{"name": "Some One", "email": "some.one@an.email.org"}],
        "docs": {
            "info_url": "http://concatenate.readthefakedocs.io",
        },
    }

    app = App.load(test_spec)

    app.make(build_dir=tmp_path / "build-dir", use_local_packages=True)

    volume_mount = str(dataset_dir) + ":/dataset:rw"
    args = [
        "/dataset",
        "--parameter",
        "dummy",
        "1",
        "--parameter",
        "print_version",
        "True",
        "--output",
        "stdout",
        OUTPUT_COL_NAME,
    ]

    dc = docker.from_env()
    try:
        dc.containers.run(
            app.reference,
            command=args,
            stderr=True,
            volumes=[volume_mount],
            user=f"{os.getuid()}:{os.getgid()}",
        )
    except docker.errors.ContainerError as e:
        raise RuntimeError(
            f"'docker run -v {volume_mount} {app.reference} {' '.join(args)}' errored:\n"
            + e.stderr.decode("utf-8")
        )

    dataset = FileSystem().load_frameset(dataset_dir)

    def strip_ver_timestamp(ver_str):
        parts = str(ver_str).split("+")
        try:
            parts[1] = parts[1].split(".")[0]
        except IndexError:
            pass
        return "+".join(parts).strip()

    assert str(dataset[OUTPUT_COL_NAME][SAMPLE_INDEX]).split(",")[0] == PACKAGE_NAME


def test_add_resources(tmp_path):

    img = P2AImage(
        name="test-resource-add-image",
        version="1.0",
        packages={
            "system": ["vim", "git"],  # just to test it out
            "pip": {
                "pydra2app": None,
            },  # just to test out the
        },
        base_image={
            "name": "python",
            "tag": "3.12.5-slim-bookworm",
            "python": "python3",
            "package_manager": "apt",
            "conda_env": None,
        },
        resources={
            "a-resource": "/internal/path/to/a/resource.txt",
            "another-resource": "/internal/path/to/another/resource",
        },
    )

    foo_file = tmp_path / "resources" / "foo.txt"
    foo_file.parent.mkdir(parents=True)
    foo_file.write_text("foo")

    resources_dir = tmp_path / "resources"
    another_resource_sub_dir = resources_dir / "another-resource"
    another_resource_sub_dir.mkdir(parents=True)
    (another_resource_sub_dir / "bar.txt").write_text("bar")

    img.make(
        build_dir=tmp_path / "build-dir",
        use_local_packages=True,
        resources={
            "a-resource": foo_file,
        },
        resources_dir=resources_dir,
    )

    dc = docker.from_env()
    args = ["cat", "/internal/path/to/a/resource.txt"]
    try:
        result = dc.containers.run(
            img.reference,
            command=args,
            stderr=True,
        )
    except docker.errors.ContainerError as e:
        raise RuntimeError(
            f"'docker run {img.reference} {' '.join(args)}' errored:\n"
            + e.stderr.decode("utf-8")
        )

    assert result == b"foo"

    args = ["cat", "/internal/path/to/another/resource/bar.txt"]
    try:
        result = dc.containers.run(
            img.reference,
            command=args,
            stderr=True,
        )
    except docker.errors.ContainerError as e:
        raise RuntimeError(
            f"'docker run {img.reference} {' '.join(args)}' errored:\n"
            + e.stderr.decode("utf-8")
        )
    assert result == b"bar"


def test_multi_command(
    simple_dataset_blueprint: TestDatasetBlueprint, tmp_path: Path
) -> None:

    dataset = simple_dataset_blueprint.make_dataset(
        FileSystem(), tmp_path / "dataset", name=""
    )

    two_dup_spec = {
        "name": "concatenate",
        "task": "frametree.testing.tasks:Concatenate",
        "row_frequency": simple_dataset_blueprint.axes.default().tostr(),
        "configuration": {"duplicates": 2},
    }

    three_dup_spec = deepcopy(two_dup_spec)
    three_dup_spec["configuration"]["duplicates"] = 3

    test_spec = {
        "name": "test_multi_commands",
        "title": "a test image for multi-image commands",
        "commands": {
            "two_duplicates": two_dup_spec,
            "three_duplicates": three_dup_spec,
        },
        "version": "1.0",
        "packages": {
            "system": ["vim", "git"],  # just to test it out
            "pip": {
                "fileformats": None,
                "pydra": None,
                "pydra2app": None,
                "frametree": None,
            },
        },
        "authors": [{"name": "Some One", "email": "some.one@an.email.org"}],
        "docs": {
            "info_url": "http://concatenate.readthefakedocs.io",
        },
    }

    app = App.load(test_spec)

    app.make(build_dir=tmp_path / "build-dir", use_local_packages=True)

    volume_mount = str(dataset.id) + ":/dataset:rw"
    base_args = [
        "/dataset",
        "--input",
        "in_file1",
        "file1",
        "--input",
        "in_file2",
        "file2",
        "--output",
        "out_file",
    ]

    fnames = ["file1.txt", "file2.txt"]

    for command in ["two_duplicates", "three_duplicates"]:

        # Name the output column based on the command and set the command
        args = base_args + [command + "_sink", "--command", command]

        dc = docker.from_env()
        try:
            dc.containers.run(
                app.reference,
                command=args,
                stderr=True,
                volumes=[volume_mount],
                user=f"{os.getuid()}:{os.getgid()}",
            )
        except docker.errors.ContainerError as e:
            raise RuntimeError(
                f"'docker run -v {volume_mount} {app.reference} {' '.join(args)}' errored:\n"
                + e.stderr.decode("utf-8")
            )

        # Add source column to saved dataset
        reloaded = dataset.reload()
        sink = reloaded[command + "_sink"]
        duplicates = 2 if command == "two_duplicates" else 3
        expected_contents = "\n".join(fnames * duplicates)
        for item in sink:
            with open(item) as f:
                contents = f.read()
            assert contents == expected_contents


def test_serialization_roundtrip(tmp_path: Path) -> None:

    test_spec = {
        "name": "native_python_test",
        "title": "a test image spec",
        "commands": {
            "python-test-command": {
                "task": {
                    "type": "shell",
                    "executable": [
                        "pydra2app",
                        "--version<version>",
                    ],
                    "inputs": {
                        "dummy": {
                            "type": int | None,
                            "help": "not actually used",
                            "argstr": None,  # won't be printed to the command line
                        }
                    },
                },
                "row_frequency": "common:Samples[sample]",
            },
        },
        "version": "1.0",
        "packages": {
            "system": ["git"],  # just to test it out
            "pip": {
                "pydra2app": None,
                "frametree": None,
                "pydra": None,
            },  # just to test out the
        },
        "base_image": {
            "name": "python",
            "tag": "3.12.5-slim-bookworm",
            "python": "python3",
            "package_manager": "apt",
            "conda_env": None,
        },
        "authors": [{"name": "Some One", "email": "some.one@an.email.org"}],
        "docs": {
            "info_url": "http://concatenate.readthefakedocs.io",
        },
    }

    app = App.load(test_spec)

    # Serialize the app to a file
    save_path = tmp_path / (app.name + ".json")
    app.save(save_path)
    reloaded_app = App.load(save_path)
    assert app.commands[0] == reloaded_app.commands[0]
    assert app == reloaded_app



ORG = "pydra2app-tests"
REGISTRY = "a.docker.registry.io"
IMAGE_VERSION = "1.0"


LICENSE_CONTENTS = "license contents"

LICENSE_NAME = "testlicense"

LICENSE_INPUT_FIELD = "expected_license_contents"

LICENSE_OUTPUT_FIELD = "validated_license_file"

LICENSE_PATH_PARAM = "expected_license_path"

LICENSE_INPUT_PATH = "contents-file"

LICENSE_OUTPUT_PATH = "validated-file"



class Pydra2AppException(Exception):
    @property
    def msg(self):
        return self.args[0]

    @msg.setter
    def msg(self, msg):
        self.args = (msg,) + self.args[1:]


class Pydra2AppError(Pydra2AppException):
    pass


class Pydra2AppRuntimeError(Pydra2AppError):
    pass


class Pydra2AppNotBoundToAnalysisError(Pydra2AppError):
    pass


class Pydra2AppVersionError(Pydra2AppError):
    pass


class Pydra2AppRequirementNotFoundError(Pydra2AppVersionError):
    pass


class Pydra2AppVersionNotDetectableError(Pydra2AppVersionError):
    pass


class Pydra2AppEnvModuleNotLoadedError(Pydra2AppError):
    pass


class Pydra2AppMissingInputError(Pydra2AppException):
    pass


class Pydra2AppProtectedOutputConflictError(Pydra2AppError):
    pass


class Pydra2AppCantPickleAnalysisError(Pydra2AppError):
    pass


class Pydra2AppRepositoryError(Pydra2AppError):
    pass


class Pydra2AppUsageError(Pydra2AppError):
    pass


class Pydra2AppCacheError(Pydra2AppError):
    pass


class Pydra2AppDesignError(Pydra2AppError):
    pass


class NamedPydra2AppError(Pydra2AppError):
    def __init__(self, name, msg):
        super(NamedPydra2AppError, self).__init__(msg)
        self.name = name


class Pydra2AppNameError(NamedPydra2AppError):
    pass


class Pydra2AppWrongFrequencyError(NamedPydra2AppError):
    pass


class Pydra2AppIndexError(Pydra2AppError):
    def __init__(self, index, msg):
        super(Pydra2AppIndexError, self).__init__(msg)
        self.index = index


class Pydra2AppDataMatchError(Pydra2AppUsageError):
    pass


class Pydra2AppPipelinesStackError(Pydra2AppError):
    pass


class Pydra2AppMissingDataException(Pydra2AppPipelinesStackError):
    pass


class Pydra2AppOutputNotProducedException(Pydra2AppPipelinesStackError):
    """
    Raised when a given spec is not produced due to switches and inputs
    provided to the analysis
    """


class Pydra2AppInsufficientRepoDepthError(Pydra2AppError):
    pass


class Pydra2AppLicenseNotFoundError(Pydra2AppNameError):
    pass


class Pydra2AppUnresolvableFormatException(Pydra2AppException):
    pass


class Pydra2AppFileSetNotCachedException(Pydra2AppException):
    pass


class NoMatchingPipelineException(Pydra2AppException):
    pass


class Pydra2AppModulesError(Pydra2AppError):
    pass


class Pydra2AppModulesNotInstalledException(Pydra2AppException):
    pass


class Pydra2AppJobSubmittedException(Pydra2AppException):
    """
    Signifies that a pipeline has been submitted to a scheduler and
    a return value won't be returned.
    """


class Pydra2AppNoRunRequiredException(Pydra2AppException):
    """
    Used to signify when a pipeline doesn't need to be run as all
    required outputs are already present in the store
    """


class Pydra2AppFileFormatClashError(Pydra2AppError):
    """
    Used when two mismatching data formats are registered with the same
    name or extension
    """


class Pydra2AppConverterNotAvailableError(Pydra2AppError):
    "The converter required to convert between formats is not"

    "available"


class Pydra2AppReprocessException(Pydra2AppException):
    pass


class Pydra2AppWrongRepositoryError(Pydra2AppError):
    pass


class Pydra2AppIvalidParameterError(Pydra2AppError):
    pass


class Pydra2AppRequirementVersionsError(Pydra2AppError):
    pass


class Pydra2AppXnatCommandError(Pydra2AppRepositoryError):
    """
    Error in the command file used to access an XNAT repository via the XNAT
    container service.
    """


class Pydra2AppUriAlreadySetException(Pydra2AppException):
    """Raised when attempting to set the URI of an item is already set"""


class Pydra2AppDataTreeConstructionError(Pydra2AppError):
    "Error in constructing data tree by store find_rows method"


class Pydra2AppBadlyFormattedIDError(Pydra2AppDataTreeConstructionError):
    "Error attempting to extract an ID from a tree path using a user provided regex"


class Pydra2AppWrongAxesError(Pydra2AppError):
    "Provided row_frequency is not a valid member of the dataset's dimensions"


class Pydra2AppNoDirectXnatMountException(Pydra2AppException):
    "Raised when attemptint to access a file-system mount for a row that hasn't been mounted directly"

    pass


class Pydra2AppEmptyDatasetError(Pydra2AppException):
    pass


class Pydra2AppBuildError(Pydra2AppError):
    pass


class NamedError(Exception):
    def __init__(self, name, msg):
        super().__init__(msg)
        self.name = name


class NameError(NamedError):
    pass


class DataNotDerivedYetError(NamedError):
    pass


class DatatypeUnsupportedByStoreError(Pydra2AppError):
    """Raised when a data store doesn't support a given datatype"""

    def __init__(self, datatype, store):
        super().__init__(
            f"'{datatype.mime_like}' data types aren't supported by {type(store)} stores"
        )



from __future__ import annotations
import typing as ty
from pathlib import Path, PurePath
import json
import pkg_resources
import logging
from urllib.parse import urlparse
import re
from itertools import zip_longest
from typing_extensions import Self
import site
import attrs
from pydra2app.core import PACKAGE_NAME
from pydra2app.core.exceptions import Pydra2AppBuildError
from frametree.core.serialize import ObjectListConverter

logger = logging.getLogger("pydra2app")


@attrs.define(kw_only=True)
class BaseImage:

    DEFAULT_IMAGE = "debian"
    DEFAULT_IMAGE_TAG = "bookworm-slim"
    DEFAULT_CONDA_ENV = "pydra2app"
    DEFAULT_USER = "root"

    name: str = attrs.field(default=DEFAULT_IMAGE)
    tag: str = attrs.field()
    package_manager: str = attrs.field()
    python: str = attrs.field(default=None)
    conda_env: str = attrs.field()
    user: str = attrs.field(default=DEFAULT_USER)

    @property
    def reference(self) -> str:
        if self.tag:
            reference = f"{self.name}:{self.tag}"
        else:
            reference = self.name
        return reference

    @name.validator
    def name_validator(self, _: attrs.Attribute[str], name: str) -> None:
        if name == "alpine":
            raise ValueError(
                "Neurodocker (the package used to build the images) does not currently "
                "support alpine base images"
            )

    @tag.default
    def tag_default(self) -> ty.Optional[str]:
        if self.name == self.DEFAULT_IMAGE:
            tag = self.DEFAULT_IMAGE_TAG
        else:
            tag = None
        return tag

    @package_manager.default
    def package_manager_default(self) -> ty.Optional[str]:
        if self.name in ("ubuntu", "debian"):
            package_manager = "apt"
        elif self.name in ("fedora", "centos"):
            package_manager = "yum"
        else:
            package_manager = None
        return package_manager

    @package_manager.validator
    def package_manager_validator(
        self, _: attrs.Attribute[ty.Optional[str]], package_manager: ty.Optional[str]
    ) -> None:
        if package_manager is None:
            raise ValueError(
                f"Package manager must be supplied explicitly for unknown base image "
                f"'{self.name}' (note only 'apt' and 'yum' package managers are "
                "currently supported)"
            )
        if package_manager not in ("yum", "apt"):
            raise ValueError(
                f"Unsupported package manager '{package_manager}' provided. Only 'apt' "
                "and 'yum' package managers are currently supported by Neurodocker"
            )

    @conda_env.default
    def conda_env_default(self) -> ty.Optional[str]:
        if self.python:
            conda_env = None
        else:
            conda_env = self.DEFAULT_CONDA_ENV
        return conda_env


@attrs.define
class ContainerAuthor:

    name: str
    email: str
    affliation: ty.Optional[str] = None


@attrs.define
class KnownIssue:

    description: str
    url: ty.Optional[str] = None


@attrs.define
class Docs:

    info_url: str = attrs.field()
    description: ty.Optional[str] = None
    known_issues: ty.List[KnownIssue] = attrs.field(
        factory=list,
        converter=ObjectListConverter(KnownIssue),  # type: ignore[misc]
        metadata={"serializer": ObjectListConverter.aslist},
    )

    @info_url.validator
    def info_url_validator(self, _: attrs.Attribute[str], info_url: str) -> None:
        parsed = urlparse(info_url)
        if not parsed.scheme or not parsed.netloc:
            raise ValueError(
                f"Could not parse info url '{info_url}', please include URL scheme"
            )


def optional_path_converter(value: ty.Optional[str]) -> ty.Optional[Path]:
    return Path(value) if value is not None else None


@attrs.define
class License:
    """Specification of a software license that needs to be present in the container
    when the command is run.

    Parameters
    ----------
    name : str
        a name to refer to the license with. Must be unique among the licenses used
        pipelines applied to a dataset, ideally for a site. Typically named closely
        after the package it is used for along with a version number if the license,
        needs to be updated with new versions e.g. freesurfer, fsl, matlab_v2022a etc...
    destination : PurePath
        destination within the container to install the license
    description : str
        a short description of the license and what it is used for
    info_url : str
        link to website with information about license, particularly how to download it
    source : Path, optional
        path to the location of a valid license file
    store_in_image : bool
        whether the license can be stored in the image or not according to the license
        conditions
    """

    name: str = attrs.field()
    destination: PurePath = attrs.field(converter=PurePath)
    description: str = attrs.field()
    info_url: str = attrs.field()
    source: ty.Optional[Path] = attrs.field(
        default=None, converter=optional_path_converter
    )
    store_in_image: bool = False

    @info_url.validator
    def info_url_validator(self, _: attrs.Attribute[str], info_url: str) -> None:
        parsed = urlparse(info_url)
        if not parsed.scheme or not parsed.netloc:
            raise ValueError(
                f"Could not parse info url '{info_url}', please include URL scheme"
            )

    # FIXME: this doesn't work inside images
    # @source.validator
    # def source_validator(self, _, source):
    #     if source is not None and not source.exists():
    #         raise ValueError(
    #             f"Source file for {self.name} license, '{str(source)}', does not exist"
    #         )

    @classmethod
    def column_path(self, name: str) -> str:
        """The column name (and resource name) for the license if it is to be downloaded
        from the source dataset"""
        return name + self.COLUMN_SUFFIX + "@"

    COLUMN_SUFFIX = "_LICENSE"


def base_package_version_converter(v: str) -> str:
    return str(v) if v is not None else None


@attrs.define
class BasePackage:

    name: str
    version: str = attrs.field(default=None, converter=base_package_version_converter)


def pip_package_extras_converter(
    extras: ty.Union[str, ty.Iterable[str]],
) -> ty.List[str]:
    if isinstance(extras, str):
        extras = extras.split(",")
    return list(extras)


@attrs.define
class PipPackage(BasePackage):
    """Specification of a Python package"""

    url: ty.Optional[str] = None
    file_path: ty.Optional[str] = None
    extras: ty.List[str] = attrs.field(
        factory=list, converter=pip_package_extras_converter
    )

    @classmethod
    def unique(
        cls, pip_specs: ty.Iterable[PipPackage], remove_pydra2app: bool = False
    ) -> ty.List[PipPackage]:
        """Merge a list of Pip install specs so each package only appears once

        Parameters
        ----------
        pip_specs : ty.Iterable[PipPackage]
            the pip specs to merge
        remove_pydra2app : bool
            remove pydra2app if present from the merged list

        Returns
        -------
        list[PipPackage]
            the merged pip specs

        Raises
        ------
        Pydra2AppError
            if there is a mismatch between two entries of the same package
        """
        dct: ty.Dict[str, PipPackage] = {}
        for pip_spec in pip_specs:
            if isinstance(pip_spec, dict):
                pip_spec = PipPackage(**pip_spec)
            if pip_spec.name == PACKAGE_NAME and remove_pydra2app:
                continue
            try:
                prev_spec = dct[pip_spec.name]
            except KeyError:
                dct[pip_spec.name] = pip_spec
            else:
                if (
                    prev_spec.version != pip_spec.version
                    or prev_spec.url != pip_spec.url
                    or prev_spec.file_path != pip_spec.file_path
                ):
                    raise RuntimeError(
                        f"Cannot install '{pip_spec.name}' due to conflict "
                        f"between requested versions, {pip_spec} and {prev_spec}"
                    )
                prev_spec.extras.extend(pip_spec.extras)
        return list(dct.values())

    def local_package_location(self, pypi_fallback: bool = False) -> PipPackage:
        """Detect the installed locations of the packages, including development
        versions.

        Parameters
        ----------
        package : [PipPackage]
            the packages (or names of) the versions to detect
        pypi_fallback : bool, optional
            Fallback to PyPI version if requested version isn't installed locally

        Returns
        -------
        PipPackage
            the pip specification for the installation location of the package
        """
        try:
            pkg = next(
                p for p in pkg_resources.working_set if p.project_name == self.name
            )
        except StopIteration:
            if pypi_fallback:
                logger.info(
                    f"Did not find local installation of package {self.name} "
                    "falling back to installation from PyPI"
                )
                return self
            raise Pydra2AppBuildError(
                f"Did not find {self.name} in installed working set:\n"
                + "\n".join(
                    sorted(
                        p.key + "/" + p.project_name for p in pkg_resources.working_set
                    )
                )
            )
        if (
            self.version
            and (
                not (pkg.version.endswith(".dirty") or self.version.endswith(".dirty"))
            )
            and pkg.version != self.version
        ):
            msg = (
                f"Requested package {self.name}=={self.version} does "
                "not match installed " + pkg.version
            )
            if pypi_fallback:
                logger.warning(msg + " falling back to installation from PyPI")
                return self
            raise Pydra2AppBuildError(msg)
        if pkg.location is None:
            raise Pydra2AppBuildError(
                f"Could not find location of package {self.name} in installed working set, "
                f"{pkg} has no local location"
            )
        pkg_loc = Path(pkg.location).resolve()
        # Determine whether installed version of requirement is locally
        # installed (and therefore needs to be copied into image) or can
        # be just downloaded from PyPI
        if pkg_loc not in site_pkg_locs:
            # Copy package into Docker image and instruct pip to install from
            # that copy
            local_spec = PipPackage(
                name=self.name, file_path=str(pkg_loc), extras=self.extras
            )
        else:
            # Check to see whether package is installed via "direct URL" instead
            # of through PyPI
            direct_url_path = Path(pkg.egg_info) / "direct_url.json"
            if direct_url_path.exists():
                with open(direct_url_path) as f:
                    url_spec = json.load(f)
                url = url_spec["url"]
                vcs_info = url_spec.get(
                    "vcs_info", url_spec
                )  # Fallback to trying to find VCS info in the base url-spec dict
                if url.startswith("file://"):
                    local_spec = PipPackage(
                        name=self.name,
                        file_path=url[len("file://") :],
                        extras=self.extras,
                    )
                else:
                    vcs_info = url_spec.get("vcs_info", url_spec)
                    if "vcs" in vcs_info:
                        url = vcs_info["vcs"] + "+" + url
                    if "commit_id" in vcs_info:
                        url += "@" + vcs_info["commit_id"]
                    local_spec = PipPackage(name=self.name, url=url, extras=self.extras)
            else:
                local_spec = PipPackage(
                    name=self.name, version=pkg.version, extras=self.extras
                )
        return local_spec


@attrs.define
class SystemPackage(BasePackage):

    pass


@attrs.define
class CondaPackage(BasePackage):

    pass

    # REQUIRED = ["numpy", "traits"]  # FIXME: Not sure if traits is actually required


@attrs.define
class NeurodockerTemplate:

    name: str
    version: str
    args: ty.Dict[str, ty.Any] = attrs.field(factory=dict)


def python_package_converter(
    packages: ty.List[ty.Union[str, ty.Dict[str, ty.Any]]],
) -> ty.List[PipPackage]:
    """
    Split out and merge any extras specifications (e.g. "pydra2app[test]")
    between dependencies of the same package
    """
    return PipPackage.unique(
        ObjectListConverter(PipPackage)(
            packages,
        ),
        remove_pydra2app=True,
    )


@attrs.define
class Packages:

    system: ty.List[SystemPackage] = attrs.field(
        factory=list,
        converter=ObjectListConverter(SystemPackage),  # type: ignore[misc]
        metadata={"serializer": ObjectListConverter.asdict},
    )
    pip: ty.List[PipPackage] = attrs.field(
        factory=list,
        converter=python_package_converter,
        metadata={"serializer": ObjectListConverter.asdict},
    )
    conda: ty.List[CondaPackage] = attrs.field(
        factory=list,
        converter=ObjectListConverter(CondaPackage),  # type: ignore[misc]
        metadata={"serializer": ObjectListConverter.asdict},
    )
    neurodocker: ty.List[NeurodockerTemplate] = attrs.field(
        factory=list,
        converter=ObjectListConverter(NeurodockerTemplate),  # type: ignore[misc]
        metadata={"serializer": ObjectListConverter.asdict},
    )


site_pkg_locs = [Path(p).resolve() for p in site.getsitepackages()]


@attrs.define
class Resource:

    name: str
    path: Path  # the path to the resource within the container
    description: str = ""


@attrs.define
class Version:

    release: ty.Union[ty.Tuple[int, ...], str]
    suffix_label: str = ""
    suffix_number: int = 0

    @classmethod
    def parse(cls, version: ty.Union[str, Self]) -> Self:
        if not isinstance(version, str):
            if not isinstance(version, cls):
                raise ValueError(
                    f"Cannot parse version from object of type {type(version)}"
                )
            return version
        match = cls.version_re.match(version)
        if match is None:
            return cls(version)
        release = match.group("release")
        try:
            release = tuple(int(r) for r in release.split("."))
        except ValueError:
            pass
        suffix_label = match.group("suffix_l") or ""
        if suffix_label and suffix_label not in cls.SUFFIX_LABELS:
            raise ValueError(
                f"Invalid suffix label {suffix_label}, must be one of {cls.SUFFIX_LABELS}"
            )
        suffix_number = int(match.group("suffix_n")) if match.group("suffix_n") else 0
        return cls(release, suffix_label, suffix_number)

    @classmethod
    def tostr(cls, obj: Self, **kwargs: ty.Any) -> str:
        return str(obj)

    def compare(self, other: "Version") -> int:
        if (isinstance(self.release, str) and isinstance(other.release, tuple)) or (
            isinstance(self.release, tuple) and isinstance(other.release, str)
        ):
            raise ValueError("Cannot compare versions with different release types")

        if isinstance(self.release, tuple) and isinstance(other.release, tuple):
            for s, o in zip_longest(self.release, other.release, fillvalue=0):
                if s < o:
                    return -1
                if s > o:
                    return 1
        else:
            if self.release < other.release:  # type: ignore[operator]
                return -1
            if self.release > other.release:  # type: ignore[operator]
                return 1
        if self.suffix_label == "post" and other.suffix_label != "post":
            return 1
        if self.suffix_label != "post" and other.suffix_label == "post":
            return -1
        if not self.suffix_label and other.suffix_label:
            return 1
        if self.suffix_label and not other.suffix_label:
            return -1
        if self.suffix_label:
            if not other.suffix_label:
                return 1
            label_index = self.SUFFIX_LABELS.index(self.suffix_label)
            other_label_index = self.SUFFIX_LABELS.index(other.suffix_label)
            if label_index < other_label_index:
                return -1
            if label_index > other_label_index:
                return 1
            if self.suffix_number < other.suffix_number:
                return -1
            if self.suffix_number > other.suffix_number:
                return 1
        elif other.suffix_label:
            return -1
        return 0

    def __str__(self) -> str:
        release_str = (
            ".".join(str(r) for r in self.release)
            if isinstance(self.release, tuple)
            else self.release
        )
        return release_str + (
            f"-{self.suffix_label}{self.suffix_number}" if self.suffix_label else ""
        )

    def __lt__(self, other: "Version") -> bool:
        return self.compare(other) < 0

    def __le__(self, other: "Version") -> bool:
        return self.compare(other) <= 0

    def __eq__(self, other: object) -> bool:
        if isinstance(other, str):
            try:
                other_version = Version(other)
            except ValueError:
                return False
        elif not isinstance(other, Version):
            return False
        else:
            other_version = other
        return self.compare(other_version) == 0

    def __ne__(self, other: object) -> bool:
        return not (self.release == other)

    def __gt__(self, other: "Version") -> bool:
        return self.compare(other) > 0

    def __ge__(self, other: "Version") -> bool:
        return self.compare(other) >= 0

    def __repr__(self) -> str:
        return f"Version({str(self)})"

    def __hash__(self) -> int:
        return hash(str(self))

    @classmethod
    def latest(cls, versions: ty.List[ty.Union[str, "Version"]]) -> "Version":
        version_objs = [
            v if isinstance(v, cls) else cls.parse(v) for v in versions  # type: ignore[arg-type]
        ]
        return sorted(version_objs)[-1]

    def bump_postfix(self) -> "Version":
        suffix_label = self.suffix_label if self.suffix_label else "post"
        suffix_number = self.suffix_number + 1
        return Version(self.release, suffix_label, suffix_number)

    SUFFIX_LABELS = ["alpha", "beta", "rc", "post"]

    version_re = re.compile(
        (
            r"^v?(?P<release>[a-zA-Z0-9_\.]+)"
            r"(?P<suffix>-(?P<suffix_l>"
            + "|".join(SUFFIX_LABELS)
            + r")(?P<suffix_n>[0-9]+)?)?$"
        ),
        re.VERBOSE | re.IGNORECASE,
    )



from __future__ import annotations
import typing as ty
from pathlib import Path
from itertools import chain
import re
import logging
import shutil
import attrs
import yaml
import toml
from neurodocker.reproenv import DockerRenderer
from pydra2app.core import __version__
from frametree.core.serialize import (
    ObjectConverter,
    ObjectListConverter,
    ClassResolver,
)
from typing_extensions import Self
from fileformats.core import DataType
from frametree.core.axes import Axes
from frametree.core.exceptions import FrametreeCannotSerializeDynamicDefinitionError
from pydra.utils.typing import is_optional, optional_type
from pydra2app.core.utils import is_relative_to
from ..command.base import ContainerCommand
from .base import P2AImage
from .components import ContainerAuthor, License, Docs, PipPackage


logger = logging.getLogger("pydra2app")


@attrs.define(kw_only=True)
class App(P2AImage):
    """A container image that contains a command with specific inputs and outputs to run.

    Parameters
    ----------
    name : str
        a name of the app to be used in naming of packages/images
    title : str
        a single line description of the app to be exposed in UIs
    version : str
        version of the package/build of the app
    org : str
        the organisation the image will be tagged within
    base_image : BaseImage, optional
        the base image to build from
    packages : Packages, optional
        the package manager used to install system packages (should match OS on base image)
    registry : str, optional
        the container registry the image is to be installed at
    authors : list[ContainerAuthor]
        list of authors of the package
    command : ContainerCommand
        description of the command that is to be run within the image
    licenses : list[License], optional
        specification of licenses required by the commands in the container. Each dict
        should contain the 'name' of the license and the 'destination' it should be
        installed inside the container.
    docs : Docs
        information for automatically generated documentation
    loaded_from : Path
        the file the spec was loaded from, if applicable
    """

    SUBPACKAGE = "deploy"

    title: str
    authors: ty.List[ContainerAuthor] = attrs.field(
        converter=ObjectListConverter(ContainerAuthor),  # type: ignore[misc]
        metadata={"serializer": ObjectListConverter.asdict},
    )
    licenses: ty.List[License] = attrs.field(
        factory=list,
        converter=ObjectListConverter(License),  # type: ignore[misc]
        metadata={"serializer": ObjectListConverter.asdict},
    )
    docs: Docs = attrs.field(converter=ObjectConverter(Docs))  # type: ignore[misc]
    commands: ty.List[ContainerCommand] = attrs.field(
        converter=ObjectListConverter(ContainerCommand)  # type: ignore[misc]
    )
    loaded_from: Path = attrs.field(
        default=None, metadata={"asdict": False}, eq=False, hash=False
    )
    pydra2app_version: str = __version__

    @commands.validator
    def _validate_commands(
        self,
        attribute: attrs.Attribute[ty.List[ContainerCommand]],
        commands: ty.List[ContainerCommand],
    ) -> None:
        if not commands:
            raise ValueError("At least one command must be defined within that app")

    def __attrs_post_init__(self) -> None:
        # Set back-references to this image in the command spec
        for command in self.commands:
            command.image = self

    def command(self, name: ty.Optional[str] = None) -> ContainerCommand:
        if name is None:
            command = self.commands[0]  # Default to the first command
        else:
            try:
                command = next(c for c in self.commands if c.name == name)
            except StopIteration:
                raise KeyError(f"No command with name '{name}' found")
        return command

    def add_entrypoint(self, dockerfile: DockerRenderer, build_dir: Path) -> None:
        dockerfile.entrypoint(
            self.activate_conda() + ["pydra2app", "pipeline-entrypoint"]
        )

    def construct_dockerfile(
        self,
        build_dir: Path,
        use_local_packages: bool = False,
        pypi_fallback: bool = False,
        pydra2app_install_extras: ty.Sequence[str] = (),
        resources: ty.Optional[ty.Dict[str, Path]] = None,
        resources_dir: ty.Optional[Path] = None,
        **kwargs: ty.Any,
    ) -> DockerRenderer:
        """Constructs a dockerfile that wraps a with dependencies

        Parameters
        ----------
        build_dir : Path
            Path to the directory the Dockerfile will be written into copy any local
            files to
        **kwargs
            Passed onto the P2AImage.construct_dockerfile() method

        Returns
        -------
        DockerRenderer
            Neurodocker Docker renderer to construct dockerfile from
        """

        dockerfile = super().construct_dockerfile(
            build_dir=build_dir,
            use_local_packages=use_local_packages,
            pypi_fallback=pypi_fallback,
            pydra2app_install_extras=pydra2app_install_extras,
            resources=resources,
            resources_dir=resources_dir,
            **kwargs,
        )

        self.install_licenses(
            dockerfile,
            build_dir,
        )

        self.make_work_dir(dockerfile)

        self.add_entrypoint(dockerfile, build_dir)

        return dockerfile

    def install_licenses(
        self,
        dockerfile: DockerRenderer,
        build_dir: Path,
    ) -> None:
        """Generate Neurodocker instructions to install licenses within the container
        image

        Parameters
        ----------
        dockerfile : DockerRenderer
            the neurodocker renderer to append the install instructions to
        build_dir : Path
            path to build dir
        """
        # Copy licenses into build directory
        license_build_dir = build_dir / "licenses"
        license_build_dir.mkdir()
        for lic in self.licenses:
            if lic.store_in_image:
                if lic.source:
                    build_path = license_build_dir / lic.name
                    shutil.copyfile(lic.source, build_path)
                    dockerfile.copy(
                        source=[str(build_path.relative_to(build_dir))],
                        destination=str(lic.destination),
                    )
                else:
                    logger.warning(
                        "License file for '%s' was not provided, will attempt to download "
                        "from '%s' dataset-level column or site-wide license dataset at "
                        "runtime",
                        lic.name,
                        lic.column_path(lic.name),
                    )

    def make_work_dir(self, dockerfile: DockerRenderer) -> None:
        """Creates the internal work directory for the container and permissives sets
        permissions"""
        dockerfile.run("mkdir -p /wl")
        dockerfile.run("chmod 777 /wl")

    @classmethod
    def load(  # type: ignore[override]
        cls,
        yml: ty.Union[Path, ty.Dict[str, ty.Any]],
        root_dir: ty.Optional[Path] = None,
        license_paths: ty.Optional[ty.Dict[str, Path]] = None,
        licenses_to_download: ty.Optional[ty.Set[str]] = None,
        default_axes: ty.Optional[ty.Type[Axes]] = None,
        source_packages: ty.Sequence[Path] = (),
        **kwargs: ty.Any,
    ) -> Self:
        """Loads a deploy-build specification from a YAML file

        Parameters
        ----------
        yml : Path or dict
            path to the YAML file to load or loaded dictionary
        root_dir : Path, optional
            path to the root directory from which a tree of specs are being loaded from.
            The name of the root directory is taken to be the organisation the image
            belongs to, and all nested directories above the YAML file will be joined by
            '.' and prepended to the name of the loaded spec.
        license_paths : dict[str, Path], optional
            Licenses that are provided at build time to be included in the image.
        licenses_to_download : set[str], optional
            Licenses that are to be downloaded at runtime. If `license_paths` is not
            None (i.e. how to access required licenses are to be specified) then required
            licenses that are not in license_paths need to be explicitly listed in
            `licenses_to_download` otherwise an error is raised
        default_axes : type[Axes]
            the default data space to assume when one isn't explicitly defined
        source_packages : Sequence[Path]
            Paths to source packages to include in the image, will be used to determine
            the local version of the package to install
        **kwargs
            additional keyword arguments that override/augment the values loaded from
            the spec file

        Returns
        -------
        Self
            The loaded spec object
        """

        if isinstance(yml, str):
            yml = Path(yml)
        if root_dir is None:
            root_dir = Path.cwd()
        if isinstance(yml, Path):
            yml_dict = cls._load_yaml(yml)
            if not isinstance(yml_dict, dict):
                raise ValueError(f"{yml!r} didn't contain a dict!")
            if is_relative_to(yml, root_dir):
                rel_parts = yml.relative_to(root_dir).parent.parts + (yml.stem,)
                if "name" not in yml_dict:
                    yml_dict["name"] = ".".join(rel_parts[1:])

                if "org" not in yml_dict:
                    yml_dict["org"] = rel_parts[0]
            else:
                yml_dict["name"] = yml.stem

            yml_dict["loaded_from"] = yml.absolute()
        else:
            yml_dict = yml

        yml_dict.pop("type", None)  # Remove "type" from dict if present

        # Override/augment loaded values from spec
        yml_dict.update(kwargs)

        # If data-space is not defined, default to `default_axes`
        commands = yml_dict["commands"]
        if isinstance(commands, dict):
            commands = commands.values()
        for cmd in commands:
            if (
                "row_frequency" in cmd
                and re.match(r"\w+", cmd["row_frequency"])
                and default_axes
            ):
                cmd["row_frequency"] = default_axes[cmd["row_frequency"]]
            if isinstance(cmd["task"], dict) and cmd["task"].get("type") == "python":
                cmd["task"]["function"] = ClassResolver.fromstr(cmd["task"]["function"])
        image = cls(**yml_dict)

        # Replace any pip packages with local source packages
        for source_package in source_packages:
            pyproject_toml = source_package / "pyproject.toml"
            package_name = None
            if pyproject_toml.exists():
                config = toml.load(pyproject_toml)
                package_name = config.get("project", {}).get("name")
            if package_name is None:
                package_name = source_package.name
            new_pip_pkg = None
            for pip_pkg in image.packages.pip:
                if pip_pkg.name == package_name:
                    new_pip_pkg = PipPackage(
                        name=package_name,
                        file_path=str(source_package),
                        extras=pip_pkg.extras,
                    )
                    image.packages.pip.remove(pip_pkg)
                    image.packages.pip.append(new_pip_pkg)
                    break
            logger.info(
                "Could not find package %s in the pip packages of the "
                "image spec, omitting from build. Found packages: %s",
                package_name,
                '", "'.join(str(p) for p in image.packages.pip),
            )

        # Explicitly override directive in loaded spec to store license in the image

        if license_paths is not None:
            for lic in image.licenses:
                if licenses_to_download and lic.name in licenses_to_download:
                    lic.store_in_image = False
                if lic.store_in_image:
                    try:
                        lic.source = license_paths[lic.name]
                    except KeyError:
                        raise RuntimeError(
                            f"{lic.name} license has not been provided when it is "
                            "specified to be stored in the image"
                        )

        return image

    @classmethod
    def load_tree(
        cls, spec_path: Path, root_dir: ty.Optional[Path] = None, **kwargs: ty.Any
    ) -> ty.List[Self]:
        """Walk the given directory structure and load all specs found within it

        Parameters
        ----------
        spec_path : Path
            Path to spec or directory tree containing specs of the pipelines to build
        root_dir : Path
            path to the base of the spec directory
        """
        spec_path = Path(spec_path)
        if root_dir is None:
            if spec_path.is_file():
                raise ValueError(
                    "If a single file is provided, the root_dir must be specified"
                )
            logger.info("No root directory specified, using the spec directory")
            root_dir = spec_path
        root_dir = Path(root_dir)
        if spec_path.is_file():
            return [cls.load(spec_path, root_dir=root_dir, **kwargs)]
        specs = []
        for path in chain(spec_path.rglob("*.yml"), spec_path.rglob("*.yaml")):
            if not any(p.startswith(".") for p in path.parts):
                logging.info("Found container image specification file '%s'", path)
                specs.append(cls.load(path, root_dir=root_dir, **kwargs))

        return specs

    def autodoc(self, doc_dir: Path, flatten: bool) -> None:
        header = {
            "title": self.name,
            "weight": 10,
        }

        if self.org is None:
            raise ValueError(
                f"Organisation must be defined to generate documentation from App spec: {self}"
            )

        if self.loaded_from:
            header["source_file"] = str(self.loaded_from)

        if flatten:
            out_dir = doc_dir
        else:
            assert isinstance(doc_dir, Path)

            out_dir = doc_dir / self.org

            assert doc_dir in out_dir.parents or out_dir == doc_dir

            out_dir.mkdir(parents=True, exist_ok=True)

        with open(f"{out_dir}/{self.name}.md", "w") as f:
            f.write("---\n")
            yaml.dump(header, f)
            f.write("\n---\n\n")

            f.write("## Package Info\n")
            tbl_info = MarkdownTable(f, "Key", "Value")
            tbl_info.write_row("Name", self.name)
            tbl_info.write_row("Title", self.title)
            tbl_info.write_row("Version", str(self.version))
            tbl_info.write_row("Base image", escaped_md(self.base_image.reference))
            tbl_info.write_row(
                "Maintainer", f"{self.authors[0].name} ({self.authors[0].email})"
            )
            tbl_info.write_row("Info URL", self.docs.info_url)

            for known_issue in self.docs.known_issues:
                tbl_info.write_row(
                    "Known issues",
                    (
                        known_issue.description + f" ({known_issue.url})"
                        if known_issue.url
                        else ""
                    ),
                )

            desc = self.docs.description or self.title
            f.write(f"\n{desc}\n\n")

            if self.licenses:
                f.write("### Required licenses\n")

                tbl_lic = MarkdownTable(f, "Name", "URL", "Description")
                for lic in self.licenses:
                    tbl_lic.write_row(
                        lic.name,
                        escaped_md(lic.info_url),
                        lic.description.strip(),
                    )

                f.write("\n")

            f.write("## Commands\n")

            for command in self.commands:

                tbl_cmd = MarkdownTable(f, "Key", "Value")

                # if command.configuration is not None:
                #     config = command.configuration
                #     # configuration keys are variable depending on the workflow class
                try:
                    class_address = ClassResolver.tostr(command.task)
                except FrametreeCannotSerializeDynamicDefinitionError:
                    class_address = command.task.__name__
                tbl_cmd.write_row("Task", class_address)
                freq_name = (
                    command.row_frequency.name
                    if not isinstance(command.row_frequency, str)
                    else re.match(r".*\[(\w+)\]", command.row_frequency).group(1)
                )
                tbl_cmd.write_row("Operates on", freq_name)

                f.write("#### Inputs\n")
                tbl_inputs = MarkdownTable(
                    f,
                    "Name",
                    "Data-type(s)",
                    "Required",
                    "Description",
                )
                for inpt in command.input_fields:
                    tbl_inputs.write_row(
                        escaped_md(inpt.name),
                        self._data_format_html(inpt.type),
                        "Y" if inpt.mandatory else "N",
                        inpt.help,
                    )
                f.write("\n")

                f.write("#### Outputs\n")
                tbl_outputs = MarkdownTable(
                    f,
                    "Name",
                    "Data-type(s)",
                    "Always generated",
                    "Description",
                )
                for outpt in command.output_fields:
                    tbl_outputs.write_row(
                        escaped_md(outpt.name),
                        self._data_format_html(optional_type(outpt.type)),
                        "Y" if not is_optional(outpt.type) else "N",
                        outpt.help,
                    )
                f.write("\n")

                if command.parameters is not None:
                    f.write("#### Parameters\n")
                    tbl_params = MarkdownTable(
                        f, "Name", "Data-type(s)", "Default", "Description"
                    )
                    for param in command.parameter_fields:
                        tbl_params.write_row(
                            escaped_md(param.name),
                            self._data_format_html(param.type),
                            escaped_md(param.default if not param.mandatory else "-"),
                            param.help,
                        )
                    f.write("\n")

    # @classmethod
    # def load_in_image(cls, spec_path: Path = IN_DOCKER_SPEC_PATH):
    #     yml_dct = cls._load_yaml(spec_path)
    #     klass = ClassResolver(cls)(yml_dct.pop("type"))
    #     return klass.load(yml_dct)

    @classmethod
    def _data_format_html(cls, datatype: ty.Union[str, DataType]) -> str:
        datatype_str = datatype.mime_like if not isinstance(datatype, str) else datatype

        return (
            f'<span data-toggle="tooltip" data-placement="bottom" title="{datatype_str}" '
            f'aria-label="{datatype_str}">{datatype_str}</span>'
        )

    DOCKERFILE_README_TEMPLATE = """
        The following Docker image was generated by Pydra2App v{} to enable the
        commands to be run in the XNAT container service. See
        https://raw.githubusercontent.com/Australian-Imaging-Service/pydra2app/main/LICENSE
        for licence.

        {}

        """


class MarkdownTable:
    def __init__(self, f: ty.IO[str], *headers: str) -> None:
        self.headers = tuple(headers)

        self.f = f
        self._write_header()

    def _write_header(self) -> None:
        self.write_row(*self.headers)
        self.write_row(*("-" * len(x) for x in self.headers))

    def write_row(self, *cols_tple: str) -> None:
        cols = list(cols_tple)
        if len(cols) > len(self.headers):
            raise ValueError(
                f"More entries in row ({len(cols)} than columns ({len(self.headers)})"
            )

        # pad empty column entries if there's not enough
        cols += [""] * (len(self.headers) - len(cols))

        # TODO handle new lines in col
        self.f.write(
            "|" + "|".join(str(col).replace("|", "\\|") for col in cols) + "|\n"
        )


def escaped_md(value: str) -> str:
    if not value:
        return ""
    return f"`{value}`"



from __future__ import annotations
import typing as ty
from pathlib import PurePath, Path, PosixPath
import json
import re
import tempfile
import requests
import itertools
from enum import Enum
from functools import cached_property
import logging
from copy import copy
import shutil
from inspect import isclass, isfunction
from build import ProjectBuilder
import attrs
import yaml
import docker.errors
from deepdiff import DeepDiff
from typing_extensions import Self
from neurodocker.reproenv import DockerRenderer
from pydra2app.core import __version__
from pydra2app.core import PACKAGE_NAME
from frametree.core.serialize import (
    ClassResolver,
    ObjectConverter,
    ObjectListConverter,
)
from frametree.core.axes import Axes
from pydra2app.core.utils import (
    DOCKER_HUB,
    GITHUB_CONTAINER_REGISTRY,
    extract_file_from_docker_image,
)
from pydra2app.core.exceptions import Pydra2AppBuildError
from .components import Packages, BaseImage, PipPackage, Resource, Version

logger = logging.getLogger("pydra2app")


@attrs.define(kw_only=True, auto_attribs=False)
class P2AImage:
    """
    Base class with Pydra2App installed within it from which all container image
    specifications can inherit from

    name : str
        name of the package/pipeline
    version : Version
        version of the package/pipeline
    org : str
        the organisation the image will be tagged within
    base_image : pydra2app.core.image.components.BaseImage, optional
        the base image to build from
    packages : Packages
        System (OS), PyPI, Conda and Neurodocker packages/templates to be installed
        in the image
    registry : str, optional
        the container registry the image is to be installed at
    readme : str, optional
        text to include in the image README file
    labels : dict[str, str]
        labels to add into the built image
    schema_version : str
        the version of the specification language used to define the image (i.e. this)
    """

    IN_DOCKER_FRAMETREE_HOME_DIR = "/frametree-home"
    IN_DOCKER_SPEC_PATH = "/pydra2app-spec.yaml"
    SCHEMA_VERSION = "2.0"
    PIP_DEPENDENCIES: ty.Tuple[str, ...] = ()
    DEFAULT_PYTHON_VERSION = "3.11"

    name: str = attrs.field()
    version: Version = attrs.field(
        converter=Version.parse, metadata={"serializer": Version.tostr}  # type: ignore[misc]
    )
    base_image: BaseImage = attrs.field(
        default=BaseImage(), converter=ObjectConverter(BaseImage)  # type: ignore[misc]
    )
    packages: Packages = attrs.field(
        default=Packages(), converter=ObjectConverter(Packages)  # type: ignore[misc]
    )
    resources: ty.List[Resource] = attrs.field(
        factory=list,
        converter=ObjectListConverter(Resource),  # type: ignore[misc]
        metadata={"serializer": ObjectListConverter.asdict},
    )
    org: ty.Optional[str] = attrs.field(default=None)
    registry: str = attrs.field(default=DOCKER_HUB)
    readme: ty.Optional[str] = attrs.field(default=None)
    labels: ty.Optional[ty.Dict[str, str]] = attrs.field(default=None)
    schema_version: str = attrs.field(default=SCHEMA_VERSION)

    @property
    def reference(self) -> str:
        return f"{self.path}:{self.tag}"

    @property
    def tag(self) -> str:
        return str(self.version)

    @property
    def path(self) -> str:
        prefix = self.registry + "/" if self.registry != DOCKER_HUB else ""
        org_str = self.org + "/" if self.org else ""
        return (prefix + org_str + self.name).lower()

    def make(
        self,
        build_dir: ty.Optional[Path] = None,
        generate_only: bool = False,
        no_cache: bool = False,
        stream_output: ty.Optional[bool] = None,
        reference: ty.Optional[str] = None,
        **kwargs: ty.Any,
    ) -> None:
        """Makes the container image from the spec: generates the Dockerfile and then
        builds it.

        Parameters
        ----------
        build_dir : Path, optional
            _description_, by default None
        generate_only: bool, optional
            whether to only generate the Dockerfile and not build the image, by default
            False
        no_cache : bool, optional
            whether to cache the build layers or not, by default False
        stream_output : bool, optional
            whether to stream the output of the build process to stdout as it is being
            built. If None, the output will be streamed if the logger level is set to
            INFO or lower, by default None
        reference : str, optional
            the tag to assign to the built image, by default the one constructed from
            the image spec will be used
        **kwargs : Any
            additional keyword arguments to pass to the Dockerfile constructor
        """

        if build_dir is None:
            build_dir = Path(tempfile.mkdtemp())
        elif not isinstance(build_dir, Path):
            build_dir = Path(build_dir)
        if build_dir.exists():
            shutil.rmtree(build_dir)
        build_dir.mkdir()

        dockerfile = self.construct_dockerfile(build_dir, **kwargs)

        image_reference = reference if reference is not None else self.reference

        if generate_only:
            self.write_dockerfile(dockerfile, build_dir, image_reference)
        else:
            self.build(
                dockerfile,
                build_dir,
                image_reference=image_reference,
                no_cache=no_cache,
                stream_output=stream_output,
            )

    def registry_tags(self) -> ty.List[str]:
        """List all the tags for the given docker image reference within the registry"""
        tags: ty.List[str]
        if self.registry == DOCKER_HUB:
            url = f"https://hub.docker.com/v2/repositories/{self.org}/{self.name}/tags/"
            tags = []
            while url:
                response = requests.get(url)
                if response.status_code == 404:
                    return []
                elif response.status_code != 200:
                    response.raise_for_status()
                data = response.json()
                tags.extend(tag["name"] for tag in data["results"])
                url = data["next"]  # Get the URL for the next page of results
        elif self.registry == GITHUB_CONTAINER_REGISTRY:
            token_uri = f"https://{self.registry}/token?scope=repository:{self.org}/{self.name}:pull"
            response = requests.get(token_uri)
            if response.status_code == 404:
                return []
            elif response.status_code != 200:
                response.raise_for_status()
            registry_token = response.json().get("token")
            url = f"https://{self.registry}/v2/{self.org}/{self.name}/tags/list"
            headers = {
                "Accept": "application/vnd.github.v3+json",
                "Authorization": f"Bearer {registry_token}",
            }
            response = requests.get(url, headers=headers)
            if response.status_code != 200:
                response.raise_for_status()
            tags = response.json().get("tags", [])
        else:
            protocol = "http" if self.registry.startswith("localhost") else "https"
            url = f"{protocol}://{self.registry}/v2/{self.org}/{self.name}/tags/list"
            response = requests.get(url)
            if response.status_code == 404:
                return []
            elif response.status_code != 200:
                response.raise_for_status()
            tags = response.json().get("tags", [])
        return tags

    @cached_property
    def latest_published(self) -> ty.Optional[Version]:
        """Get the latest published version of the image from the registry"""
        published_tags = self.registry_tags()
        logger.info(
            "Found the following versions in the registry for '%s': %s",
            self.path,
            published_tags,
        )
        versions = sorted(Version.parse(t) for t in published_tags)
        return versions[-1] if versions else None

    def matches_image(self, image_reference: str) -> bool:
        """Check if the specifications of two images match"""
        try:
            extracted_specs_file = extract_file_from_docker_image(
                image_reference, PosixPath(self.IN_DOCKER_SPEC_PATH)
            )
        except docker.errors.NotFound:
            extracted_specs_file = None
        if extracted_specs_file is None:
            logger.warning(
                "Could not extract spec from existing image '%s' to check for changes "
                "assuming it is different",
                image_reference,
            )
            return False
        logger.info(
            "Comparing build spec with that of existing image %s",
            image_reference,
        )
        built_spec = self.load(extracted_specs_file)

        changelog = self.compare_specs(built_spec, check_versions=False)

        if changelog:
            logger.debug(
                "'%s' differs from existing image '%s':\n%s",
                self.reference,
                image_reference,
                changelog,
            )
            return False
        return True

    def construct_dockerfile(
        self,
        build_dir: Path,
        use_local_packages: bool = False,
        pypi_fallback: bool = False,
        pydra2app_install_extras: ty.Sequence[str] = (),
        resources: ty.Optional[ty.Dict[str, Path]] = None,
        resources_dir: ty.Optional[Path] = None,
        **kwargs: ty.Any,
    ) -> DockerRenderer:
        """Constructs a dockerfile that wraps a with dependencies

        Parameters
        ----------
        build_dir : Path
            Path to the directory the Dockerfile will be written into copy any local
            files to
        use_local_packages: bool, optional
            Use the python package versions that are installed within the
            current environment, i.e. instead of pulling from PyPI. Useful during
            development and testing
        pypi_fallback : bool, optional
            whether to fallback to packages installed on PyPI when versions of
            local packages don't match installed
        pydra2app_install_extras : Iterable[str], optional
            Extras for the Pydra2App package that need to be installed into the
            dockerfile (e.g. tests)
        resources : dict[str, Path], optional
            Resources to be copied into the docker image, keys of the dictionary should
            match resource names defined in the pipeline specification, values of the
            dictionary should be paths to the resources on the local filesystem to be
            copied into the docker image
        resources_dir : Path, optional
            Alternative to supplying the resources separately, a directory containing
            the resources to be copied into the docker image as named subdirectories
            can be provided

        Returns
        -------
        DockerRenderer
            Neurodocker Docker renderer to construct dockerfile from
        """

        if not build_dir.is_dir():
            raise Pydra2AppBuildError(
                f"Build dir '{str(build_dir)}' is not a valid directory"
            )

        build_dir = build_dir.absolute()

        dockerfile = self.init_dockerfile()

        dockerfile.user("root")

        self.install_system_packages(dockerfile)

        self.install_package_templates(dockerfile)

        self.install_python(
            dockerfile,
            build_dir,
            use_local_packages=use_local_packages,
            pypi_fallback=pypi_fallback,
            pydra2app_install_extras=pydra2app_install_extras,
        )

        self.add_resources(dockerfile, build_dir, resources, resources_dir)

        self.write_readme(dockerfile, build_dir)

        self.add_labels(dockerfile)

        self.insert_spec(dockerfile, build_dir)

        # Create writable directories
        for dpath in (self.IN_DOCKER_FRAMETREE_HOME_DIR, "/.cache"):
            dockerfile.run(f"mkdir {dpath}")
            dockerfile.run(f"chmod 777 {dpath}")
        dockerfile.env(FRAMETREE_HOME=self.IN_DOCKER_FRAMETREE_HOME_DIR)

        return dockerfile

    @classmethod
    def write_dockerfile(
        cls,
        dockerfile: DockerRenderer,
        build_dir: Path,
        image_reference: str,
    ) -> None:
        """Writes the dockerfile to the specified build directory

        Parameters
        ----------
        dockerfile : DockerRenderer
            Neurodocker renderer to build
        build_dir : Path
            path of the build directory
        image_reference : str
            Docker image tag to assign to the built image
        """
        # Save generated dockerfile to file
        out_file = build_dir / "Dockerfile"
        out_file.parent.mkdir(exist_ok=True, parents=True)
        with open(str(out_file), "w") as f:
            f.write(dockerfile.render())
        logger.info(
            "Dockerfile for '%s' generated at %s", image_reference, str(out_file)
        )

    @classmethod
    def build(
        cls,
        dockerfile: DockerRenderer,
        build_dir: Path,
        image_reference: str,
        no_cache: bool = False,
        stream_output: ty.Optional[bool] = None,
    ) -> str:
        """Builds the dockerfile in the specified build directory

        Parameters
        ----------
        dockerfile : DockerRenderer
            Neurodocker renderer to build
        build_dir : Path
            path of the build directory
        image_reference : str
            Docker image tag to assign to the built image
        no_cache : bool, optional
            whether to cache the build layers or not, by default False
        stream_output : bool, optional
            whether to stream the output of the build process to stdout as it is being
            built. If None, the output will be streamed if the logger level is set to
            INFO or lower, by default None

        Returns
        -------
        str
            the image ID of the built image

        Raises
        ------
        docker.errors.BuildError
            If the build process fails
        """
        if stream_output is None:
            stream_output = logger.level <= logging.INFO

        cls.write_dockerfile(dockerfile, build_dir, image_reference)
        dc = docker.from_env()

        response = dc.api.build(
            path=str(build_dir.absolute()), tag=image_reference, rm=True, decode=True
        )
        last_event = None
        result_stream, progress_stream = itertools.tee(response)
        for chunk in progress_stream:
            if "stream" in chunk:
                if stream_output:
                    print(chunk["stream"], end="")
                match = re.search(
                    r"(^Successfully built |sha256:)([0-9a-f]+)$", chunk["stream"]
                )
                if match:
                    logging.info("Successfully built docker image %s", image_reference)
                    return match.group(2)
            if "error" in chunk:
                raise docker.errors.BuildError(
                    chunk["error"],
                    (
                        f"Building '{image_reference}' from '{str(build_dir)}/Dockerfile': "
                        + str(result_stream)
                    ),
                )
            last_event = chunk
        raise docker.errors.BuildError(last_event or "Unknown", result_stream)

    def init_dockerfile(self) -> DockerRenderer:
        dockerfile = DockerRenderer(self.base_image.package_manager).from_(
            self.base_image.reference
        )
        return dockerfile

    def save(self, yml_path: Path) -> None:
        """Saves the specification to a YAML file that can be loaded again

        Parameters
        ----------
        yml_path : Path
            path to file to save the spec to
        """
        yml_dct = self.asdict()
        yml_dct["type"] = ClassResolver.tostr(self, strip_prefix=False)
        with open(yml_path, "w") as f:
            yaml.dump(yml_dct, f)

    def insert_spec(self, dockerfile: DockerRenderer, build_dir: Path) -> None:
        """Generate Neurodocker instructions to save the specification inside the built
        image to be used when running the command and comparing against future builds

        Parameters
        ----------
        dockerfile : DockerRenderer
            the neurodocker renderer to append the install instructions to
        spec : dict
            the specification used to build the image
        build_dir : Path
            path to build dir
        """
        self.save(build_dir / "pydra2app-spec.yaml")
        dockerfile.copy(
            source=["./pydra2app-spec.yaml"], destination=self.IN_DOCKER_SPEC_PATH
        )

    @classmethod
    def load(
        cls,
        yml: ty.Union[Path, ty.Dict[str, ty.Any]],
        name: ty.Optional[str] = None,
        **kwargs: ty.Any,
    ) -> Self:
        """Loads a deploy-build specification from a YAML file

        Parameters
        ----------
        yml : Path or dict
            path to the YAML file to load or loaded dictionary
        name: str, optional
            name of the pipeline, by default None

        Returns
        -------
        Self
            The loaded spec object
        """

        if isinstance(yml, str):
            yml = Path(yml)
        if isinstance(yml, Path):
            yml_dict = cls._load_yaml(yml)
            if not isinstance(yml_dict, dict):
                raise ValueError(f"{yml!r} didn't contain a dict!")
            if name is None:
                name = yml.stem
            yml_dict["loaded_from"] = yml.absolute()
        else:
            yml_dict = yml
        if yml_dict.get("name") is None:
            yml_dict["name"] = name

        yml_dict.pop("type", None)  # Remove "type" from dict if present

        # Override/augment loaded values from spec
        yml_dict.update(kwargs)

        return cls(**yml_dict)

    def add_resources(
        self,
        dockerfile: DockerRenderer,
        build_dir: Path,
        resources: ty.Optional[ty.Dict[str, Path]],
        resources_dir: ty.Optional[Path],
    ) -> None:
        """Add static resources to the docker image"""
        if resources_dir is not None:
            all_resources = {
                p.name: p
                for p in resources_dir.iterdir()
                if p.is_dir() and not p.name.startswith(".")
            }
        else:
            all_resources = {}
        if resources:
            all_resources.update(resources)
        resources_dir_context = build_dir / "resources"
        resources_dir_context.mkdir(exist_ok=True)
        for resource in self.resources:
            try:
                local_path = all_resources[resource.name]
            except KeyError:
                resource_dir_str = (
                    str(resources_dir) if resources_dir is not None else None
                )
                raise RuntimeError(
                    f"Resource '{resource.name}' specified in the pipeline specification "
                    "but not provided in the 'resources' argument or a sub-directory "
                    f"of 'resources_dir' ({resource_dir_str!r})\n"
                    + "\n".join(all_resources)
                )
            # copy local path into Docker build dir so it is included in context
            build_context_path = resources_dir_context / resource.name
            if local_path.is_dir():
                shutil.copytree(local_path, build_context_path)
            else:
                shutil.copy(local_path, build_context_path)
            dockerfile.copy(
                source=[str(build_context_path.relative_to(build_dir))],
                destination=resource.path,
            )

    def add_labels(
        self, dockerfile: DockerRenderer, labels: ty.Optional[ty.Dict[str, str]] = None
    ) -> None:
        if labels is None:
            labels = self.labels
        if labels:
            dockerfile.labels({k: json.dumps(v).strip('"') for k, v in labels.items()})

    def install_python(
        self,
        dockerfile: DockerRenderer,
        build_dir: Path,
        use_local_packages: bool = False,
        pydra2app_install_extras: ty.Sequence[str] = (),
        pypi_fallback: bool = False,
    ) -> None:
        """Generate Neurodocker instructions to install an appropriate version of
        Python and the required Python packages

        Parameters
        ----------
        dockerfile : DockerRenderer
            the neurodocker renderer to append the install instructions to
        build_dir : Path
            the path to the build directory
        pydra2app_install_extras : Iterable[str]
            Optional extras (i.e. as defined in "extras_require" in setup.py) required
            for the pydra2app package
        use_local_packages: bool, optional
            Use the python package versions that are installed within the
            current environment, i.e. instead of defaulting to the release from PyPI.
            Useful during development and testing
        pydra2app_install_extras : list[str]
            list of "install extras" (options) to specify when installing Pydra2App
            (e.g. 'test')
        pypi_fallback : bool, optional
            Whether to fall back to PyPI version when local version doesn't match
            requested

        Returns
        -------
        ty.List[ty.List[str, ty.List[str, str]]]
            neurodocker instructions to install python and required packages
        """

        pip_specs = PipPackage.unique(
            self.packages.pip
            + [PipPackage(PACKAGE_NAME, extras=pydra2app_install_extras)]
            + [PipPackage(d) for d in self.PIP_DEPENDENCIES]
        )

        pip_strs = []
        local_pip_specs = []
        for pip_spec in pip_specs:
            if use_local_packages:
                pip_spec = pip_spec.local_package_location(pypi_fallback=pypi_fallback)
            if pip_spec.file_path:
                local_pip_specs.append(pip_spec)
            else:
                pip_strs.append(self.pip_spec2str(pip_spec, dockerfile, build_dir))

        conda_pkg_names = set(p.name for p in self.packages.conda)
        conda_strs: ty.List[str] = []
        # for pkg_name in CondaPackage.REQUIRED:
        #     if pkg_name not in conda_pkg_names:
        #         conda_strs.append(pkg_name)

        conda_strs.extend(
            f"{p.name}={p.version}" if p.version is not None else p.name
            for p in self.packages.conda
        )

        if not self.base_image.python:
            if "python" not in conda_pkg_names:
                conda_strs.append(f"python=={self.DEFAULT_PYTHON_VERSION}")
            conda_pip_strs = pip_strs
        else:
            conda_pip_strs = []

        if conda_strs:
            dockerfile.add_registered_template(
                "miniconda",
                version="latest",
                env_name=self.base_image.conda_env,
                env_exists=False,
                conda_install=" ".join(conda_strs),
                pip_install=" ".join(conda_pip_strs),
            )
        activate_conda = self.activate_conda() if self.base_image.conda_env else []
        if pip_strs and self.base_image.python:
            dockerfile.run(
                " ".join(
                    activate_conda
                    + [self.base_image.python, "-m", "pip", "install"]
                    + pip_strs
                )
            )
        # Local pip packages in a separate layer to avoid having to re-install
        # everything if a local package changes (e.g. during development)
        if local_pip_specs:
            python_cmd = self.base_image.python if self.base_image.python else "python3"
            dockerfile.run(
                " ".join(
                    activate_conda
                    + [python_cmd, "-m", "pip", "install"]
                    + [
                        self.pip_spec2str(s, dockerfile, build_dir)
                        for s in local_pip_specs
                    ]
                )
            )

    def install_system_packages(self, dockerfile: DockerRenderer) -> None:
        """Generate Neurodocker instructions to install systems packages in dockerfile

        Parameters
        ----------
        dockerfile : DockerRenderer
            the neurodocker renderer to append the install instructions to
        system_packages : Iterable[str]
            the packages to install on the operating system
        """
        pkg_strs = [
            f"{p.name}={p.version}" if p.version else p.name
            for p in self.packages.system
        ]
        if pkg_strs:
            dockerfile.install(pkg_strs)

    def install_package_templates(
        self,
        dockerfile: DockerRenderer,
    ) -> None:
        """Install custom packages from Neurodocker package_templates

        Parameters
        ----------
        dockerfile : DockerRenderer
            the neurodocker renderer to append the install instructions to
        package_templates : Iterable[ty.Dict[str, str]]
            Neurodocker installation package_templates to be installed inside the image. A
            dictionary containing the 'name' and 'version' of the template along
            with any additional keyword arguments required by the template
        """
        for template in self.packages.neurodocker:
            kwds = attrs.asdict(template)
            # so we can pop the name and opt args and leave the original dictionary intact
            kwds = copy(kwds)
            kwds.update(kwds.pop("args", {}))
            dockerfile.add_registered_template(kwds.pop("name"), **kwds)

    @classmethod
    def pip_spec2str(
        cls,
        pip_spec: PipPackage,
        dockerfile: DockerRenderer,
        build_dir: Path,
    ) -> str:
        """Generates a string to be passed to `pip` in order to install a package
        from a "pip specification" object

        Parameters
        ----------
        pip_spec : PipPackage
            specification of the package to install
        dockerfile : DockerRenderer
            Neurodocker Docker renderer object used to generate the Dockerfile
        build_dir : Path
            path to the directory the Docker image will be built in

        Returns
        -------
        str
            string to be passed to `pip` installer
        """
        # Copy the local development versions of Python dependencies into the
        # docker image if present, instead of relying on the PyPI version,
        # which might be missing local changes and bugfixes (particularly in testing)
        # if use_local_packages:
        #     pip_spec = pip_spec.local_package_location(pypi_fallback=pypi_fallback)
        if pip_spec.file_path:
            if pip_spec.version or pip_spec.url:
                raise Pydra2AppBuildError(
                    "Cannot specify a package by `file_path`, `version` and/or " "`url`"
                )
            # pkg_build_path = cls.copy_sdist_into_build_dir(
            #     pip_spec.file_path, build_dir
            # )
            # Create a source distribution tarball to be installed within the docker
            # image
            sdist_dir = build_dir / cls.PYTHON_PACKAGE_DIR
            builder = ProjectBuilder(pip_spec.file_path)
            pkg_build_path = Path(builder.build("sdist", sdist_dir))
            pip_str = "/" + cls.PYTHON_PACKAGE_DIR + "/" + pkg_build_path.name
            dockerfile.copy(
                source=[str(pkg_build_path.relative_to(build_dir))], destination=pip_str
            )
        elif pip_spec.url:
            if pip_spec.version:
                raise Pydra2AppBuildError(
                    "Cannot specify a package by `url` and `version`"
                )
            pip_str = pip_spec.url
        else:
            pip_str = pip_spec.name
        if pip_spec.extras:
            pip_str += "[" + ",".join(pip_spec.extras) + "]"
        if pip_spec.version:
            pip_str += "==" + pip_spec.version
        return pip_str

    # @classmethod
    # def copy_sdist_into_build_dir(cls, local_installation: Path, build_dir: Path):
    #     """Create a source distribution from a locally installed "editable" python package
    #     and copy it into the build dir so it can be installed in the Docker image

    #     Parameters
    #     ----------
    #     package_name : str
    #         the name of the package (how it will be called in the docker image)
    #     local_installation : Path
    #         path to the local installation
    #     build_dir : Path
    #         path to the build directory

    #     Returns
    #     -------
    #     Path
    #         the path to the source distribution within the build directory
    #     """
    #     sdist_dir = build_dir / cls.PYTHON_PACKAGE_DIR
    #     built = build_package(local_installation, sdist_dir, ["sdist"])
    #     return sdist_dir / built[0]

    def write_readme(self, dockerfile: DockerRenderer, build_dir: Path) -> None:
        """Generate Neurodocker instructions to install README file inside the docker
        image

        Parameters
        ----------
        dockerfile : DockerRenderer
            the neurodocker renderer to append the install instructions to
        description : str
            a description of what the pipeline does, to be inserted in a README file
            in the Docker image
        build_dir : Path
            path to build dir
        """
        with open(build_dir / "README.md", "w") as f:
            f.write(self.DOCKERFILE_README_TEMPLATE.format(__version__, self.readme))
        dockerfile.copy(source=["./README.md"], destination="/README.md")

    def asdict(self) -> ty.Dict[str, ty.Any]:
        """Return a serialized version of the pipeline image specification that can be
        written to file"""

        def filter(attr: attrs.Attribute[ty.Any], value: ty.Any) -> bool:
            if isinstance(value, type(self)):
                return False  # filter-out back-references to self
            if not attr.metadata.get("asdict", True):
                return False  # filter out explicitly excluded attributes
            try:
                if value == attr.default:
                    return False  # filter out default values
            except (AttributeError, TypeError):
                return True
            return True

        def listify_containers(value: ty.Any) -> ty.Any:
            if isinstance(value, ty.Mapping):
                return {k: listify_containers(v) for k, v in value.items()}
            elif isinstance(value, (ty.Sequence, set, frozenset)) and not isinstance(
                value, (str, bytes)
            ):
                return [listify_containers(v) for v in value]
            return value

        def serializer(
            _: ty.Any, attr: attrs.Attribute[ty.Any], value: ty.Any
        ) -> ty.Any:
            if attr is not None and "serializer" in attr.metadata:
                return attr.metadata["serializer"](
                    value,
                    value_serializer=serializer,
                    filter=filter,
                )
            elif isinstance(value, Axes):
                if hasattr(self, "commands") and self.commands[0].AXES:
                    return str(value)
                else:
                    return value.tostr()
            elif isinstance(value, Enum):
                return value.value
            elif isinstance(value, PurePath):
                return str(value)
            # elif isinstance(value, pydra.compose.base.Task):
            #     return unstructure(value)
            elif isclass(value) or isfunction(value) or ty.get_origin(value):
                return ClassResolver.tostr(value, strip_prefix=False)
            return listify_containers(value)

        return attrs.asdict(self, value_serializer=serializer, filter=filter)

    def activate_conda(self) -> ty.List[str]:
        """Generate the preamble to a command line that activates the conda environment

        Returns
        -------
        str
            part of a command line, which activates the conda environment
        """
        if not self.base_image.conda_env:
            return []
        return ["conda", "run", "--no-capture-output", "-n", self.base_image.conda_env]

    def compare_specs(self, other: Self, check_versions: bool = True) -> DeepDiff:
        """Compares two build specs against each other and returns the difference

        Parameters
        ----------
        s1 : dict
            first spec
        s2 : dict
            second spec
        check_version : bool
            check the pydra2app version used to generate the specs

        Returns
        -------
        DeepDiff
            the difference between the specs
        """

        sdict = self.asdict()
        odict = other.asdict()

        def prep(s: ty.Dict[str, ty.Any]) -> ty.Dict[str, ty.Any]:
            dct = {
                k: v
                for k, v in s.items()
                if (not k.startswith("_") and (v or isinstance(v, bool)))
            }
            if check_versions:
                if "pydra2app_version" not in dct:
                    dct["pydra2app_version"] = __version__
            else:
                dct.pop("pydra2app_version", None)
                dct.pop("version", None)
            return dct

        diff = DeepDiff(prep(sdict), prep(odict), ignore_order=True)
        return diff

    @classmethod
    def _load_yaml(cls, yaml_file: ty.Union[Path, str]) -> ty.Dict[str, ty.Any]:
        def yaml_join(loader: yaml.Loader, node: yaml.SequenceNode) -> str:
            seq = loader.construct_sequence(node)
            return "".join([str(i) for i in seq])

        # Add special constructors to handle joins and concatenations within the YAML
        yaml.SafeLoader.add_constructor(tag="!join", constructor=yaml_join)
        with open(yaml_file, "r") as f:
            dct = yaml.load(f, Loader=yaml.SafeLoader)
        return dct  # type: ignore[no-any-return]

    DOCKERFILE_README_TEMPLATE = """
        The following Docker image was generated by Pydra2App v{} (https://pydra2app.readthedocs.io)

        {}

        """  # noqa: E501

    PYTHON_PACKAGE_DIR = "python-packages"



from .app import App
from .tasks import define_identity, IdentityConverter, ExtractFromJson
from .helpers import value_from_stdout

__all__ = [
    "App",
    "define_identity",
    "IdentityConverter",
    "ExtractFromJson",
    "value_from_stdout",
]



from __future__ import annotations
import shutil
import re
from copy import copy
import tempfile
import json
import inspect
import logging
from pathlib import Path
import typing as ty
import sys
from collections import defaultdict
import attrs
from attrs.converters import default_if_none
import pydra.compose.base
from fileformats.core import DataType, Field
from pydra.utils import get_fields, structure, unstructure
import pydra.utils.general
from pydra.utils.typing import optional_type
from pydra.compose.base import Arg, Out
from frametree.core.exceptions import FrametreeCannotSerializeDynamicDefinitionError
from pydra.utils.typing import is_fileset_or_union
from frametree.core.serialize import ClassResolver
from frametree.core.utils import show_workflow_errors, path2label
from frametree.core.row import DataRow
from frametree.core.frameset.base import FrameSet
from frametree.core.store import Store
from frametree.core.axes import Axes
from pydra2app.core.exceptions import Pydra2AppUsageError
from pydra2app.core import PACKAGE_NAME


if ty.TYPE_CHECKING:
    from ..image import App


logger = logging.getLogger("pydra2app")

DEFAULT_TASK_NAME = "ContainerCommandTask"


def task_converter(
    task_class: str | dict[str, ty.Any],
) -> type[pydra.compose.base.Task]:

    task_cls: type[pydra.compose.base.Task]

    if isinstance(task_class, str):
        task_cls = ClassResolver(  # type: ignore[misc]
            pydra.compose.base.Task,
            alternative_types=[ty.Callable],
            package=PACKAGE_NAME,
        )(task_class)
    elif isinstance(task_class, dict):

        if task_class["type"] == "python":
            task_class["function"] = ClassResolver.fromstr(task_class["function"])

        for field_dct in list(task_class.get("inputs", {}).values()) + list(
            task_class.get("outputs", {}).values()
        ):
            if isinstance(field_dct, dict):
                type_ = field_dct.get("type", None)
                if isinstance(type_, str):
                    field_dct["type"] = ClassResolver.fromstr(type_)

        task_cls = structure(task_class)
    elif issubclass(task_class, pydra.compose.base.Task):
        task_cls = task_class
    else:
        raise TypeError(f"Cannot convert {type(task_class)} ({task_class}) to a task")
    return task_cls


def task_equals(
    task_cls: type[pydra.compose.base.Task],
) -> tuple[str, pydra.utils.general._TaskFields]:
    """Used to compare task classes to see if they are equivalent."""
    return task_cls._task_type(), get_fields(task_cls)


def task_serializer(
    task_cls: type[pydra.compose.base.Task],
    **kwargs: ty.Any,
) -> str | dict[str, ty.Any]:
    """Serializes a task to a dictionary

    Parameters
    ----------
    task : type[pydra.compose.base.Task]
        the task to serialize
    **kwargs: Any
        keyword arguments passed to the `unstructure` serializer

    Returns
    -------
    str | dict[str, ty.Any]
        the serialized task, either as a import location, or as a serialised dictionary
        of the task definition if the import location is not available (i.e. the task was
        dynamically created)
    """
    try:
        address: str = ClassResolver.tostr(task_cls, strip_prefix=False)
    except FrametreeCannotSerializeDynamicDefinitionError:
        dct: dict[str, ty.Any] = unstructure(task_cls, **kwargs)
        return dct
    else:
        return address


@attrs.define(kw_only=True, auto_attribs=False)
class ContainerCommand:
    """A definition of a command to be run within a container. A command wraps up a
    task or workflow to provide/configure a UI for convenient launching.

    Parameters
    ----------
    task : pydra.compose.base.Task or str
        the task to run or the location of the class
    row_frequency: Axes, optional
        the frequency that the command operates on
    parameters: list[str], optional
        inputs of the task to be treated as fixed parameters entered
        by the user (i.e. rather than drawn from the data store). By default,
        any non-file task input is considered a parameter, however, if a list of
        parameters is provided then only those inputs will be treated as parameters
        and any non-file inputs will be treated as fields to be pulled from metadata.
        File inputs marked as parameters will be treated as paths to local files (i.e.
        outside of the store) or URLs to be downloaded from the web depending on their
        format.
    configuration: ty.Dict[str, ty.Any]
        constant values used to configure the task/workflow, i.e. not presented to the
        user.
    image: App
        back-reference to the image the command is installed in
    """

    STORE_TYPE = "file_system"
    AXES: ty.Optional[ty.Type[Axes]] = None

    task: type[pydra.compose.base.Task] = attrs.field(
        converter=task_converter,
        metadata={"serializer": task_serializer},
        eq=task_equals,
    )
    name: str = attrs.field()
    row_frequency: ty.Optional[Axes] = attrs.field(default=None)
    configuration: ty.Dict[str, ty.Any] = attrs.field(
        factory=dict, converter=default_if_none(dict)  # type: ignore[misc]
    )
    parameters: ty.List[str] = attrs.field(converter=list)
    image: App = attrs.field(
        default=None, eq=False, hash=False, metadata={"asdict": False}
    )

    @name.default
    def _default_name(self) -> str:
        return self.task.__name__

    @parameters.default
    def _default_parameters(self) -> ty.List[str]:
        """By default, any non-file task input is considered a parameter that is not
        fixed in the configuration"""
        return [
            i.name
            for i in get_fields(self.task)
            if not (
                i.name == self.task._executor_name
                or is_fileset_or_union(i.type)
                or i.type is DataRow
                or i.name in self.configuration
                or isinstance(i, Out)
            )
        ]

    @parameters.validator
    def _validate_parameters(
        self, attribute: attrs.Attribute[ty.Any], value: ty.List[str]
    ) -> None:
        """Validates that the parameters are valid task inputs"""
        task_inputs = [i.name for i in get_fields(self.task)]
        for param in value:
            if param not in task_inputs:
                raise ValueError(
                    f"Parameter '{param}' is not a valid input to task {self.task}"
                )
            if param in self.configuration:
                raise ValueError(
                    f"Parameter '{param}' cannot be both a parameter and a configuration "
                    "argument"
                )

    @configuration.validator
    def _validate_configuration(
        self, attribute: attrs.Attribute[ty.Any], value: ty.Dict[str, ty.Any]
    ) -> None:
        """Validates that the configuration arguments are valid task inputs"""
        task_inputs = [i.name for i in get_fields(self.task)]
        for param in value:
            if param not in task_inputs:
                raise ValueError(
                    f"Configuration argument '{param}' is not a valid input to task "
                    f"{self.task}"
                )

    def __attrs_post_init__(self) -> None:
        if isinstance(self.row_frequency, Axes):
            pass
        elif isinstance(self.row_frequency, str):
            try:
                self.row_frequency = Axes.fromstr(self.row_frequency)
            except ValueError:
                if self.AXES:
                    self.row_frequency = self.AXES[self.row_frequency]
                else:
                    raise ValueError(
                        f"'{self.row_frequency}' row frequency cannot be resolved to a axes, "
                        "needs to be of form <axes>[<row-frequency>]"
                    )
        elif self.AXES:
            self.row_frequency = self.AXES.default()
        else:
            raise ValueError(
                f"Value for row_frequency must be provided to {type(self).__name__}.__init__ "
                "because it doesn't have a defined AXES class attribute"
            )

    @property
    def inputs(self) -> ty.List[str]:
        """The inputs to the task"""
        non_inputs = (
            self.parameters + list(self.configuration) + [self.task._executor_name]
        )
        return [
            f.name
            for f in get_fields(self.task)
            if f.name not in non_inputs and not isinstance(f, Out)
        ]

    @property
    def outputs(self) -> ty.List[str]:
        """The outputs of the task"""
        return [o.name for o in get_fields(self.task.Outputs)]

    @property
    def input_fields(self) -> ty.List[Arg]:
        fields = get_fields(self.task)
        return [fields[i] for i in self.inputs if fields[i].type is not DataRow]

    @property
    def output_fields(self) -> ty.List[Out]:
        fields = get_fields(self.task.Outputs)
        return [fields[i] for i in self.outputs]

    @property
    def parameter_fields(self) -> ty.List[Arg]:
        fields = get_fields(self.task)
        return [fields[p] for p in self.parameters]

    def input_field(self, name: str) -> Arg:
        if name not in self.inputs:
            raise ValueError(
                f"Input field '{name}' is not a valid input to task {self.task} "
                f"(available: {self.inputs})"
            )
        return get_fields(self.task)[name]

    def output_field(self, name: str) -> Arg:
        if name not in self.outputs:
            raise ValueError(
                f"Input field '{name}' is not a valid output of task {self.task} "
                f"(available: {self.outputs})"
            )
        return get_fields(self.task.Outputs)[name]

    def parameter_field(self, name: str) -> Arg:
        if name not in self.parameters:
            raise ValueError(
                f"Input field '{name}' is not a valid output of task {self.task} "
                f"(available: {self.parameters})"
            )
        return get_fields(self.task)[name]

    @property
    def axes(self) -> ty.Type[Axes]:
        return type(self.row_frequency)

    def configuration_args(self) -> ty.List[str]:

        # Set up fixed arguments used to configure the workflow at initialisation
        cmd_args = []
        if self.configuration is not None:
            for cname, cvalue in self.configuration.items():
                cvalue_json = json.dumps(cvalue)
                cmd_args.append(f"--configuration {cname} '{cvalue_json}' ")

        return cmd_args

    def license_args(self) -> ty.List[str]:
        cmd_args = []
        if self.image:
            for lic in self.image.licenses:
                if lic.source is None:
                    cmd_args.append(f"--download-license {lic.name} {lic.destination}")
        return cmd_args

    def execute(
        self,
        address: str,
        input_values: ty.Optional[ty.Dict[str, str]] = None,
        output_values: ty.Optional[ty.Dict[str, str]] = None,
        parameter_values: ty.Optional[ty.Dict[str, ty.Any]] = None,
        work_dir: ty.Optional[Path] = None,
        ids: ty.Union[ty.List[str], str, None] = None,
        dataset_hierarchy: ty.Optional[str] = None,
        dataset_name: ty.Optional[str] = None,
        overwrite: bool = False,
        loglevel: str = "warning",
        worker: ty.Optional[str] = None,
        export_work: ty.Optional[Path] = None,
        raise_errors: bool = False,
        keep_running_on_errors: bool = False,
        pipeline_name: ty.Optional[str] = None,
        **store_kwargs: ty.Any,
    ) -> None:
        """Runs the command within the entrypoint of the container image.

        Performs a number of steps in one long pipeline that would typically be done
        in separate command calls when running manually, i.e.:

            * Loads a dataset, creating if it doesn't exist
            * create input and output columns if they don't exist
            * applies the pipeline to the dataset
            * runs the pipeline

        Parameters
        ----------
        dataset : FrameSet
            dataset ID str (<store-nickname>//<dataset-id>:<dataset-name>)
        input_values : dict[str, str]
            values passed to the inputs of the command
        output_values : dict[str, str]
            values passed to the outputs of the command
        parameter_values : dict[str, ty.Any]
            values passed to the parameters of the command
        store_cache_dir : Path
            cache path used to download data from the store to the working node (if necessary)
        pipeline_cache_dir : Path
            cache path created when running the pipelines
        plugin : str
            Pydra plugin used to execute the pipeline
        ids : list[str]
            IDs of the dataset rows to run the pipeline over
        overwrite : bool, optional
            overwrite existing outputs
        export_work : Path
            export work directory to an alternate location after the workflow is run
            (e.g. for forensics)
        raise_errors : bool
            raise errors instead of capturing and logging (for debugging)
        pipeline_name : str
            the name to give to the pipeline, defaults to the name of the command image
        **store_kwargs: Any
            keyword args passed through to Store.load
        """
        if input_values is None:
            input_values = {}
        elif not isinstance(input_values, dict):
            input_values = dict(input_values)
        if output_values is None:
            output_values = {}
        elif not isinstance(output_values, dict):
            output_values = dict(output_values)
        if parameter_values is None:
            parameter_values = {}
        elif not isinstance(parameter_values, dict):
            parameter_values = dict(parameter_values)

        if unrecognised := set(input_values) - set(self.inputs):
            raise ValueError(
                f"Unrecognised input values passed to command {self.name}:\n"
                f"unrecognised={unrecognised}\n"
                f"available={list(self.inputs)}\n"
            )
        if unrecognised := set(output_values) - set(self.outputs):
            raise ValueError(
                f"Unrecognised output values passed to command {self.name}:\n"
                f"unrecognised={unrecognised}\n"
                f"available={list(self.outputs)}\n"
            )

        if unrecognised := set(parameter_values) - set(self.parameters):
            raise ValueError(
                f"Unrecognised parameter values passed to command {self.name}:\n"
                f"unrecognised={unrecognised}\n"
                f"available={list(self.parameters)}\n"
            )

        if missing := set(i.name for i in self.input_fields if i.mandatory) - set(
            n
            for n, v in input_values.items()
            if v or (v == "" and self.input_field(n).type is str)
        ):
            raise ValueError(
                f"Missing mandatory input values passed to command {self.name}:\n"
                f"missing={missing}\n"
            )

        if missing := set(p.name for p in self.parameter_fields if p.mandatory) - set(
            n
            for n, v in parameter_values.items()
            if v or (v == "" and self.parameter_field(n).type is str)
        ):
            raise ValueError(
                f"Missing mandatory parameter values passed to command {self.name}:\n"
                f"missing={missing}\n"
            )

        if isinstance(export_work, bytes):
            export_work = Path(export_work.decode("utf-8"))

        if loglevel != "none":
            logging.basicConfig(
                stream=sys.stdout, level=getattr(logging, loglevel.upper())
            )

        if work_dir is None:
            work_dir = Path(tempfile.mkdtemp())

        if pipeline_name is None:
            pipeline_name = self.name

        work_dir = Path(work_dir)
        work_dir.mkdir(parents=True, exist_ok=True)

        store_cache_dir = work_dir / "store-cache"
        pipeline_cache_dir = work_dir / "pydra"

        frameset = self.load_frameset(
            address, store_cache_dir, dataset_hierarchy, dataset_name, **store_kwargs
        )

        # Install required software licenses from store into container
        if self.image is not None:
            frameset.download_licenses(
                [lic for lic in self.image.licenses if not lic.store_in_image]
            )

        input_values = dict(input_values) if input_values else {}
        output_values = dict(output_values) if output_values else {}
        parameter_values = dict(parameter_values) if parameter_values else {}

        converter_args = {}  # Arguments passed to converter
        pipeline_inputs = []
        for input_name in self.inputs:
            inpt = self.input_field(input_name)
            if inpt.type is DataRow:
                pipeline_inputs.append(("frametree_data_row__", inpt.name, inpt.type))
                continue
            input_path = input_values.get(input_name, None)
            if not input_path:
                assert not inpt.mandatory, "missing " + input_name
                logger.info("No value provided for input '%s', skipping", input_name)
                continue

            path, qualifiers = self.extract_qualifiers_from_path(input_path)
            source_kwargs = qualifiers.pop("criteria", {})
            if match := re.match(r"<(\w+)(@\w+)?>", path):
                column_name = match.group(1)
                if frameset_qualifier := match.group(2):
                    source_frameset = frameset.store[frameset_qualifier[1:]]
                    column = source_frameset[column_name]
                else:
                    column = frameset[column_name]
                logger.info(f"Found existing source column {column}")
            else:
                default_column_name = f"{path2label(self.name)}_{input_name}"
                try:
                    column = frameset[default_column_name]
                except KeyError:
                    logger.info(f"Adding new source column '{default_column_name}'")
                    datatype = (
                        Field.from_primitive(inpt.type)
                        if inspect.isclass(inpt.type)
                        and not issubclass(inpt.type, DataType)
                        else inpt.type.convertible_from()
                    )  # TODO: Create a union of all the convertible datatypes for FileSet types
                    column = frameset.add_source(
                        name=default_column_name,
                        datatype=datatype,
                        path=path,
                        is_regex=True,
                        **source_kwargs,
                    )
                else:
                    logger.info("Found existing source column %s", default_column_name)

            pipeline_inputs.append((column.name, inpt.name, inpt.type))
            converter_args[column.name] = qualifiers.pop("converter", {})
            if qualifiers:
                raise Pydra2AppUsageError(
                    "Unrecognised qualifier namespaces extracted from path for "
                    f"{inpt.name} (expected ['criteria', 'converter']): {qualifiers}"
                )

        pipeline_outputs = []
        for output_name in self.outputs:
            output = self.output_field(output_name)
            output_path = output_values.get(output_name, None)
            if not output_path:
                logger.info("No value provided for output '%s', skipping", output_name)
                continue
            path, qualifiers = self.extract_qualifiers_from_path(output_path)
            if "@" not in path:
                path = f"{path}@{frameset.name}"  # Add dataset namespace
            sink_name = path2label(path)
            if sink_name in frameset.columns:
                column = frameset[sink_name]
                if not column.is_sink:
                    raise Pydra2AppUsageError(
                        f"Output column name '{sink_name}' shadows existing source column"
                    )
                logger.info(f"Found existing sink column {column}")
            else:
                logger.info(f"Adding new source column '{sink_name}'")
                datatype = (
                    Field.from_primitive(output.type)
                    if not issubclass(output.type, DataType)
                    else output.type
                )
                frameset.add_sink(
                    name=sink_name,
                    datatype=datatype,
                    path=path,
                )
            pipeline_outputs.append((sink_name, output.name, output.type))
            converter_args[sink_name] = qualifiers.pop("converter", {})
            if qualifiers:
                raise Pydra2AppUsageError(
                    "Unrecognised qualifier namespaces extracted from path for "
                    f"{output_name} (expected ['criteria', 'converter']): {qualifiers}"
                )

        # if not pipeline_outputs and task_outputs:
        #     raise ValueError(
        #         f"No output values provided to command {self} "
        #         f"(available: {list(task_outputs.keys())})"
        #     )

        frameset.save()  # Save definitions of the newly added columns

        task_kwargs = copy(self.configuration)
        for param_name, param_value in parameter_values.items():
            param = self.parameter_field(param_name)
            logger.info(
                "Parameter %s (type %s) passed value %s",
                param_name,
                param.type,
                param_value,
            )
            if param.type is not str:
                if param_value == "":
                    assert not param.mandatory
                    param_value = None
                    logger.info(
                        "Non-string parameter '%s' passed empty string, setting to None",
                        param_name,
                    )
                else:
                    # Convert field from string if necessary
                    field_type = optional_type(param.type)
                    try:
                        field_type = Field.from_primitive(field_type)
                    except TypeError:
                        pass
                    param_value = field_type(param_value)

            task_kwargs[param_name] = param_value

        task = self.task(**task_kwargs)

        if pipeline_name in frameset.pipelines and not overwrite:
            pipeline = frameset.pipelines[self.name]
            if task != pipeline.task:
                raise RuntimeError(
                    f"A pipeline named '{self.name}' has already been applied to "
                    "which differs from one specified. Please use '--overwrite' option "
                    "if this is intentional"
                )
        else:
            pipeline = frameset.apply(
                pipeline_name,
                task,
                inputs=pipeline_inputs,
                outputs=pipeline_outputs,
                row_frequency=self.row_frequency,
                overwrite=overwrite,
                converter_args=converter_args,
            )

        if isinstance(ids, str):
            ids = ids.split(",")

        # Instantiate the Pydra workflow
        wf = pipeline(ids=ids)

        # execute the workflow
        try:
            outputs = wf(cache_root=pipeline_cache_dir, worker=worker)
        except RuntimeError:
            msg = show_workflow_errors(
                pipeline_cache_dir, omit_nodes=["per_node", "main"]
            )
            logger.error(
                "Pipeline failed with errors for the following nodes:\n\n%s", msg
            )
            if raise_errors or not msg:
                raise
            else:
                errors = True
        else:
            logger.info(
                "Pipeline '%s' ran successfully for the following data rows:\n%s",
                pipeline_name,
                "\n".join(outputs.processed),
            )
            errors = False
        finally:
            if export_work:
                logger.info("Exporting work directory to '%s'", export_work)
                export_work.mkdir(parents=True, exist_ok=True)
                shutil.copytree(pipeline_cache_dir, export_work / "pydra")

        # Abort at the end after the working directory can be copied back to the
        # host so that XNAT knows there was an error
        if errors:
            if keep_running_on_errors:
                while True:
                    pass
            else:
                sys.exit(1)

    @classmethod
    def extract_qualifiers_from_path(
        cls, user_input: str
    ) -> ty.Tuple[str, ty.Dict[str, ty.Any]]:
        """Extracts out "qualifiers" from the user-inputted paths. These are
        in the form 'path ns1.arg1=val1 ns1.arg2=val2, ns2.arg1=val3...

        Parameters
        ----------
        col_name : str
            name of the column the
        user_input : str
            The path expression + qualifying keyword args to extract

        Returns
        -------
        path : str
            the path expression stripped of qualifiers
        qualifiers : defaultdict[dict]
            the extracted qualifiers
        """
        qualifiers: ty.Dict[str, ty.Any] = defaultdict(dict)
        if "=" in user_input:  # Treat user input as containing qualifiers
            parts = re.findall(r'(?:[^\s"]|"(?:\\.|[^"])*")+', user_input)
            path = parts[0].strip('"')
            for part in parts[1:]:
                try:
                    full_name, val = part.split("=", maxsplit=1)
                except ValueError as e:
                    e.args = ((e.args[0] + f" attempting to split '{part}' by '='"),)
                    raise e
                try:
                    ns, name = full_name.split(".", maxsplit=1)
                except ValueError as e:
                    e.args = (
                        (e.args[0] + f" attempting to split '{full_name}' by '.'"),
                    )
                    raise e
                try:
                    val = json.loads(val)
                except json.JSONDecodeError:
                    pass
                qualifiers[ns][name] = val
        else:
            path = user_input
        return path, qualifiers

    def load_frameset(
        self,
        address: str,
        cache_dir: Path,
        dataset_hierarchy: ty.Optional[str],
        dataset_name: ty.Optional[str],
        **kwargs: ty.Any,
    ) -> FrameSet:
        """Loads a dataset from within an image, to be used in image entrypoints

        Parameters
        ----------
        address : str
            dataset ID str
        cache_dir : Path
            the directory to use for the store cache
        dataset_hierarchy : str, optional
            the hierarchy of the dataset
        dataset_name : str
            overwrite dataset name loaded from ID str
        **kwargs: Any
            passed through to Store.load

        Returns
        -------
        _type_
            _description_
        """
        try:
            dataset = FrameSet.load(address, **kwargs)
        except KeyError:

            store_name, id, name = FrameSet.parse_id_str(address)

            if dataset_name is not None:
                name = dataset_name

            store = Store.load(store_name, cache_dir=cache_dir, **kwargs)

            if dataset_hierarchy is None:
                hierarchy = self.axes.default().span()
            else:
                hierarchy = dataset_hierarchy.split(",")

            try:
                dataset = store.load_frameset(
                    id, name
                )  # FIXME: Does this need to be here or this covered by L253??
            except KeyError:
                dataset = store.define_frameset(id, hierarchy=hierarchy, axes=self.axes)
        return dataset



import shutil
from pathlib import Path
import typing as ty
from copy import deepcopy
import attrs
from pydra.compose import python
import fileformats.core
from fileformats.generic import File
from fileformats.text import TextFile
from fileformats.testing import EncodedText
import fileformats.text
from frametree.core.row import DataRow
from pydra2app.testing.constants import LICENSE_OUTPUT_FIELD


@python.define
def add(a: float, b: float) -> float:
    return a + b


@python.define(outputs=["path", "suffix"])
def path_manip(dpath: Path, fname: str) -> ty.Tuple[str, str]:
    path = dpath / fname
    return str(path), path.suffix


@attrs.define(auto_attribs=True)
class A:
    x: int
    y: int


@attrs.define(auto_attribs=True)
class B:
    u: float
    v: float


@attrs.define(auto_attribs=True)
class C:
    z: float


@python.define(outputs=["c"])
def AttrsFunc(a: A, b: B) -> C:
    return C(z=a.x * b.u + a.y * b.v)


@python.define
def Plus10ToFileNumbers(filenumber_row: DataRow) -> None:
    """Alters the item paths within the data row, by converting them to
    an int and adding 10. Used in the test_run_pipeline_on_row_cli test.

    Parameters
    ----------
    row : DataRow
        the data row to modify
    """

    # Make a deep copy so the hash of the original isn't modified
    filenumber_row = deepcopy(filenumber_row)
    for entry in filenumber_row.entries:
        item = fileformats.text.TextFile(ty.cast(fileformats.core.FileSet, entry.item))
        new_item_stem = str(int(item.stem) + 10)
        shutil.move(item.fspath, item.fspath.parent / (new_item_stem + item.actual_ext))


@python.define(outputs=["out_file"])
def IdentityFile(in_file: File) -> File:
    return in_file


@python.define(outputs=["out_file"])
def IdentityTextFile(in_file: TextFile) -> TextFile:
    """
    Parameters
    ----------
    in_file : TextFile
        the input text file

    Returns
    -------
    out_file: TextFile
        the output text file
    """
    return in_file


@python.define(outputs=["out_file"])
def IdentityEncodedText(in_file: EncodedText) -> EncodedText:
    return in_file


@python.define
def identity(in_: ty.Any) -> ty.Any:
    return in_


@python.define
def MultiplyContents(
    in_file: File,
    multiplier: ty.Union[int, float],
    out_file: ty.Optional[Path] = None,
    dtype: type = float,
) -> File:
    """Multiplies the contents of the file, assuming that it contains numeric
    values on separate lines

    Parameters
    ----------
    in_file : Path
        path to input file to multiply the contents of
    multiplier : int or float
        the multiplier to apply to the file values
    out_file : Path
        the path to write the output file to
    dtype : type
        the type to cast the file contents to"""

    if out_file is None:
        out_file = Path("out_file.txt").absolute()

    with open(in_file) as f:
        contents = f.read()

    multiplied = []
    for line in contents.splitlines():
        multiplied.append(str(dtype(line.strip()) * multiplier))

    with open(out_file, "w") as f:
        f.write("\n".join(multiplied))

    return File(out_file)


@python.define
def ContentsAreNumeric(in_file: File) -> bool:
    """Checks the contents of a file to see whether each line can be cast to a numeric
    value

    Parameters
    ----------
    in_file : Path
        the path to a text file

    Returns
    -------
    bool
        if all the lines are numeric return True
    """
    with open(in_file) as f:
        contents = f.read()
    try:
        float(contents.strip())
    except ValueError:
        return False
    return True


@python.define(outputs=[LICENSE_OUTPUT_FIELD])
def CheckLicence(
    expected_license_path: File,
    expected_license_contents: File,
) -> File:
    """Checks the `expected_license_path` to see if there is a file with the same contents
    as that of `expected_license_contents`

    Parameters
    ----------
    expected_license_path : File
        path to the expected license file
    expected_license_contents : File
        path containing the contents expected in the expected license file

    Returns
    -------
    File
        passes through the expected license file so the task can be connected back to the
        dataset
    """
    with open(expected_license_contents) as f:
        expected_contents = f.read()
    with open(expected_license_path) as f:
        actual_contents = f.read()
    if expected_contents != actual_contents:
        raise Exception(
            f'License contents "{actual_contents}" did not match '
            f'expected "{expected_contents}"'
        )
    return expected_license_contents


TEST_TASKS = {
    "add": (add, {"a": 4, "b": 5}, {"out": 9}),
    "path_manip": (
        path_manip,
        {"dpath": Path("/home/foo/Desktop"), "fname": "bar.txt"},
        {"path": "/home/foo/Desktop/bar.txt", "suffix": ".txt"},
    ),
    "attrs_func": (
        AttrsFunc,
        {"a": A(x=2, y=4), "b": B(u=2.5, v=1.25)},
        {"c": C(z=10)},
    ),
}

BASIC_TASKS = ["add", "path_manip", "attrs_func"]

FILE_TASKS = ["concatenate"]



import typing as ty
from pathlib import Path
import random
import docker
import os
import logging
from copy import copy
from traceback import format_exc
from pydra2app.core.image import App
from pydra2app.core.image.components import Version
from pydra2app.core.utils import DOCKER_HUB, GITHUB_CONTAINER_REGISTRY
import pytest

logger = logging.getLogger("pydra2app")

VERSIONS = [
    "1.0-alpha0",
    "1.0-alpha2",
    "1.0-beta0",
    "1.0.0",
    "1.0-post1",
    "1.0-post2",
    "1.1-alpha0",
    "1.2.0",
    "1.11.0",
    "1.12.1-rc1",
]


@pytest.fixture
def image_spec(command_spec) -> ty.Dict[str, ty.Any]:
    return {
        "org": "australian-imaging-service",
        "name": "test-pipeline",
        "version": "1.0.0",
        "title": "A pipeline to test Pipeline2app's deployment tool",
        "commands": {"concatenate-test": command_spec},
        "authors": [{"name": "Thomas G. Close", "email": "thomas.close@sydney.edu.au"}],
        "docs": {
            "info_url": "http://concatenate.readthefakedocs.io",
        },
        "readme": "This is a test pipeline",
        "packages": {
            "system": ["vim", "git"],
        },
    }


REQUIRED_ENVVARS = ("GHCR_USERNAME", "GHCR_TOKEN", "DOCKER_USERNAME", "DOCKER_TOKEN")

REGISTRIES = [GITHUB_CONTAINER_REGISTRY, DOCKER_HUB, "localhost"]


@pytest.fixture(params=REGISTRIES)
def docker_registry(request: pytest.FixtureRequest, local_docker_registry: str) -> str:
    return request.param if request.param != "localhost" else local_docker_registry


@pytest.fixture
def image_tags(image_spec, docker_registry, tmp_path) -> ty.List[str]:

    registry_prefix = docker_registry.split(".")[0].upper()
    username = os.environ.get(f"{registry_prefix}_USERNAME")
    token = os.environ.get(f"{registry_prefix}_TOKEN")

    dc = docker.from_env()

    if username is not None and token is not None:
        response = dc.login(username=username, password=token, registry=docker_registry)
        if response["Status"] != "Login Succeeded":
            logger.warning("Could not login to '%s':\n\n%s", docker_registry, response)

    pushed = []

    for version in VERSIONS:
        build_dir = tmp_path / f"build-{version}"

        image_spec_cpy = copy(image_spec)

        image_spec_cpy["version"] = version
        if docker_registry == DOCKER_HUB:
            image_spec_cpy["org"] = "australianimagingservice"

        image = App(registry=docker_registry, **image_spec_cpy)

        try:
            dc.api.pull(image.reference)
        except docker.errors.APIError as e:
            if e.response.status_code in (404, 500):
                image.make(build_dir=build_dir)
                try:
                    dc.api.push(image.reference)
                except Exception:
                    pytest.skip(
                        f"Could not push '{image.reference}':\n\n{format_exc()}"
                    )
            else:
                raise
        pushed.append(image.tag)

    return sorted(pushed)


def test_sort_versions():

    rng = random.Random(42)

    shuffled = copy(VERSIONS)
    rng.shuffle(shuffled)

    sorted_versions = sorted(Version.parse(v) for v in shuffled)

    assert sorted_versions == [Version.parse(v) for v in VERSIONS]
    assert [str(v) for v in sorted_versions] == VERSIONS


UNKNOWN_VERSIONS = ["1.0.0-unknown0", "1.0.0-alpha0junk"]


@pytest.mark.parametrize("version", UNKNOWN_VERSIONS)
def test_bad_version(version):
    Version.parse(version).release = version


def test_registry_tags(
    image_tags: ty.List[str],
    tmp_path: Path,
    docker_registry: str,
    image_spec: ty.Dict[str, ty.Any],
) -> None:

    image_spec_cpy = copy(image_spec)
    if docker_registry == DOCKER_HUB:
        image_spec_cpy["org"] = "australianimagingservice"

    app = App(registry=docker_registry, **image_spec_cpy)
    assert sorted(app.registry_tags()) == sorted(image_tags)



from __future__ import annotations
from typing import Union, Dict, Tuple
import typing as ty
import shutil
import re
import yaml
import tempfile
from pathlib import Path
import attrs
import pytest
import docker
from pydra2app.core.image import App

from pydra2app.core.cli import (
    make,
    make_docs,
    bootstrap,
)
from frametree.core.utils import show_cli_trace


# @pytest.mark.xfail(reason="Need to fix a couple of things after syntax changes")
def test_deploy_make_cli(command_spec, cli_runner, work_dir):

    DOCKER_ORG = "testorg"
    DOCKER_REGISTRY = "test.registry.org"
    IMAGE_GROUP_NAME = "testpkg"

    concatenate_spec = {
        "title": "a test image spec",
        "commands": {"test-command": command_spec},
        "version": "1.0",
        "packages": {
            "system": ["vim", "git"],  # just to test it out
            "pip": {"pydra": None, "pydra2app": None},  # just to test it out
        },
        "authors": [{"name": "Some One", "email": "some.one@an.email.org"}],
        "docs": {
            "info_url": "http://concatenate.readthefakedocs.io",
        },
    }

    build_dir = work_dir / "build"
    build_dir.mkdir()
    spec_path = work_dir / DOCKER_ORG
    sub_dir = spec_path / IMAGE_GROUP_NAME
    sub_dir.mkdir(parents=True)
    with open(sub_dir / "concatenate.yml", "w") as f:
        yaml.dump(concatenate_spec, f)

    result = cli_runner(
        make,
        [
            "common:App",
            str(spec_path),
            "--build-dir",
            str(build_dir),
            "--registry",
            DOCKER_REGISTRY,
            "--loglevel",
            "warning",
            "--use-local-packages",
            "--install-extras",
            "test",
            "--raise-errors",
            "--for-localhost",
            "--dont-check-registry",
            "--stream-logs",
        ],
    )
    assert result.exit_code == 0, show_cli_trace(result)
    tag = result.output.strip().splitlines()[-1]
    assert tag == f"{DOCKER_REGISTRY}/{DOCKER_ORG}/{IMAGE_GROUP_NAME}.concatenate:1.0"

    # Clean up the built image
    dc = docker.from_env()
    dc.images.remove(tag)


# @pytest.mark.xfail(reason="Need to fix the test handle invalid docker tag name used")
def test_deploy_remake_cli(command_spec, local_docker_registry, cli_runner, run_prefix):
    """Tests the check to see whether"""

    IMAGE_GROUP_NAME = "testpkg-rebuild" + run_prefix

    def build_spec(spec, **kwargs):
        work_dir = Path(tempfile.mkdtemp())
        build_dir = work_dir / "build"
        build_dir.mkdir()
        spec_path = work_dir / "testorg"
        sub_dir = spec_path / IMAGE_GROUP_NAME
        sub_dir.mkdir(parents=True)
        with open(sub_dir / "concatenate.yml", "w") as f:
            yaml.dump(spec, f)

        result = cli_runner(
            make,
            [
                "common:App",
                str(spec_path),
                "--build-dir",
                str(build_dir),
                "--registry",
                local_docker_registry,
                "--loglevel",
                "warning",
                "--use-local-packages",
                "--install-extras",
                "test",
                "--raise-errors",
                "--check-registry",
                "--for-localhost",
            ],
            **kwargs,
        )
        return result

    concatenate_spec = {
        "title": "a test image",
        "commands": {"test-command": command_spec},
        "version": "1.0",
        "packages": {"system": ["vim", "git"], "pip": ["pydra2app"]},
        "name": "test_deploy_rebuild_cli",
        "authors": [{"name": "Some One", "email": "some.one@an.email.org"}],
        "docs": {"info_url": "http://concatenate.readthefakedocs.io"},
    }

    # Build a basic image
    result = build_spec(concatenate_spec)
    assert result.exit_code == 0, show_cli_trace(result)
    assert result.output
    tag = result.output.strip().splitlines()[-1]
    try:
        dc = docker.from_env()
        dc.api.push(tag)

        # FIXME: Need to ensure that logs are captured properly then we can test this
        # result = build_spec(concatenate_spec)
        # assert "Skipping" in result.output

        # Modify the spec so it doesn't match the original that has just been
        # built (but don't increment the version number -> image tag so there
        # is a clash)
        concatenate_spec["packages"] = {"system": ["vim", "git"]}

        result = build_spec(concatenate_spec, catch_exceptions=False)

        # Check that the image was rebuilt with an incremented tag
        assert result.exit_code == 0, show_cli_trace(result)
        rebuilt_tag = result.output.strip().splitlines()[-1]
        assert rebuilt_tag.split(":")[-1] == "1.0-post1"
        dc.images.remove(rebuilt_tag)
    finally:
        # Clean up the built images
        dc.images.remove(tag)


@attrs.define
class DocsFixture:

    yaml_src: str
    markdown: str
    licenses_to_provide: ty.List[str] = attrs.field(factory=list)


docs_fixtures = {
    "simple": DocsFixture(
        """
title: a simple app
version: &package_version '0.16.1'
authors:
  - name: author_name
    email: author@email.org
base_image:
  name: abc
  tag: *package_version
  package_manager: apt
docs:
  info_url: https://example.com
  description: >-
    a test of the YAML join functionality
commands:
  identity-task:
    task: pydra2app.testing.tasks:IdentityTextFile
    row_frequency: common:Samples[sample]
    """.strip(),
        """
---
source_file: spec.yaml
title: package.spec
weight: 10

---

## Package Info
|Key|Value|
|---|-----|
|Name|package.spec|
|Title|a simple app|
|Version|0.16.1|
|Base image|`abc:0.16.1`|
|Maintainer|author_name (author@email.org)|
|Info URL|https://example.com|

a test of the YAML join functionality

## Commands
|Key|Value|
|---|-----|
|Task|pydra2app.testing.tasks:IdentityTextFile|
|Operates on|sample|
#### Inputs
|Name|Data-type(s)|Required|Description|
|----|------------|--------|-----------|
|`in_file`|<span data-toggle="tooltip" data-placement="bottom" title="text/text-file" aria-label="text/text-file">text/text-file</span>|Y|the input text file|

#### Outputs
|Name|Data-type(s)|Always generated|Description|
|----|------------|----------------|-----------|
|`out_file`|<span data-toggle="tooltip" data-placement="bottom" title="text/text-file" aria-label="text/text-file">text/text-file</span>|Y|the output text file|

#### Parameters
|Name|Data-type(s)|Default|Description|
|----|------------|-------|-----------|
""".strip(),
    ),
    #     "full": DocsFixture(
    #         """
    # title: a more involved image spec
    # version: &package_version '0.16.1'
    # authors:
    #   - name: author_name
    #     email: author@email.org
    # base_image:
    #   name: abc
    #   tag: *package_version
    #   package_manager: yum
    # docs:
    #   info_url: https://example.com
    #   description: >-
    #     a longer description
    #   known_issues:
    #     - description: Memory overrun on large file paths
    #       url: https://github.com/myorg/mypackage/issues/644
    # packages:
    #   system:
    #     vim: 99.1
    #     git:
    #   pip:
    #     - pydra
    #     - pydra-dcm2niix
    #   neurodocker:
    #     dcm2niix: v1.0.20201102
    # licenses:
    #   freesurfer:
    #     destination: /opt/freesurfer/license.txt
    #     description: >
    #       license description
    #     info_url: http://path.to.license.provider.org/licenses
    # commands:
    #   bids-task:
    #     task: bids:bids_app
    #     inputs:
    #       T1w:
    #         configuration:
    #           path: anat/T1w
    #         datatype: medimage/nifti-gz-x
    #         help: "T1-weighted anatomical scan"
    #         column_defaults:
    #           datatype: medimage/dicom-series
    #       T2w:
    #         configuration:
    #           path: anat/T2w
    #         datatype: medimage/nifti-gz-x
    #         help: "T2-weighted anatomical scan"
    #         column_defaults:
    #           datatype: medimage/dicom-series
    #       fMRI:
    #         datatype: medimage/nifti-gz-x
    #         help: "functional MRI"
    #         configuration:
    #           path: func/bold/task=rest
    #         column_defaults:
    #           datatype: medimage/dicom-series
    #     outputs:
    #       mriqc:
    #         datatype: generic/directory
    #         help: "MRIQC output directory"
    #         configuration:
    #           path: mriqc
    #     parameters:
    #       fmriprep_flags:
    #         field: flags
    #         datatype: field/text
    #         help: description of flags param
    #     row_frequency: common:Clinical[session]
    #     configuration:
    #       executable: /usr/local/miniconda/bin/mriqc
    #       dataset: /work/bids-dataset
    #       app_output_dir: /work/bids-app-output
    #     """.strip(),
    #         """
    # ---
    # source_file: /var/folders/mz/yn83q2fd3s758w1j75d2nnw80000gn/T/tmp47_dxmyq/specs/spec.yaml
    # title: package.spec
    # weight: 10
    # ---
    # ## Package Info
    # |Key|Value|
    # |---|-----|
    # |Name|package.spec|
    # |Title|a more involved image spec|
    # |Version|0.16.1|
    # |Base image|`abc:0.16.1`|
    # |Maintainer|author_name (author@email.org)|
    # |Info URL|https://example.com|
    # |Known issues|Memory overrun on large file paths (https://github.com/myorg/mypackage/issues/644)|
    # a longer description
    # ### Required licenses
    # |Name|URL|Description|
    # |----|---|-----------|
    # |freesurfer|`http://path.to.license.provider.org/licenses`|license description|
    # ## Commands
    # |Key|Value|
    # |---|-----|
    # |Task|bids:bids_app|
    # |Operates on|session|
    # #### Inputs
    # |Name|Required data-type|Default column data-type|Description|
    # |----|------------------|------------------------|-----------|
    # |`T1w`|<span data-toggle="tooltip" data-placement="bottom" title="medimage/nifti-gz-x" aria-label="medimage/nifti-gz-x">medimage/nifti-gz-x</span>|<span data-toggle="tooltip" data-placement="bottom" title="medimage/dicom-series" aria-label="medimage/dicom-series">medimage/dicom-series</span>|T1-weighted anatomical scan|
    # |`T2w`|<span data-toggle="tooltip" data-placement="bottom" title="medimage/nifti-gz-x" aria-label="medimage/nifti-gz-x">medimage/nifti-gz-x</span>|<span data-toggle="tooltip" data-placement="bottom" title="medimage/dicom-series" aria-label="medimage/dicom-series">medimage/dicom-series</span>|T2-weighted anatomical scan|
    # |`fMRI`|<span data-toggle="tooltip" data-placement="bottom" title="medimage/nifti-gz-x" aria-label="medimage/nifti-gz-x">medimage/nifti-gz-x</span>|<span data-toggle="tooltip" data-placement="bottom" title="medimage/dicom-series" aria-label="medimage/dicom-series">medimage/dicom-series</span>|functional MRI|
    # #### Outputs
    # |Name|Required data-type|Default column data-type|Description|
    # |----|------------------|------------------------|-----------|
    # |`mriqc`|<span data-toggle="tooltip" data-placement="bottom" title="generic/directory" aria-label="generic/directory">generic/directory</span>|<span data-toggle="tooltip" data-placement="bottom" title="generic/directory" aria-label="generic/directory">generic/directory</span>|MRIQC output directory|
    # #### Parameters
    # |Name|Data type|Description|
    # |----|---------|-----------|
    # |`fmriprep_flags`|`str`|description of flags param|
    # """.strip(),
    #         ["freesurfer"],
    #     ),
}


@pytest.mark.parametrize("fixture", docs_fixtures.items(), ids=lambda x: x[0])
def test_make_docs_cli(
    cli_runner, run_prefix, work_dir: Path, fixture: Tuple[str, DocsFixture]
):
    fixture_name, fixture_content = fixture

    # TODO handle multiple 'files' in a fixture
    print(f"Processing fixture: {fixture_name!r}")
    output = _make_docs(cli_runner, work_dir, fixture_content.yaml_src)

    strip_source_file_re = re.compile(r"source_file:.*")

    stripped_output = strip_source_file_re.sub("", output)
    stripped_reference = strip_source_file_re.sub("", fixture_content.markdown)

    assert (
        stripped_output == stripped_reference
    ), f"Fixture {fixture_name!r} didn't match output"


def _make_docs(
    cli_runner,
    work_dir: Path,
    docs: Union[str, Dict[str, str]],
    *args,
    flatten: ty.Optional[bool] = None,
) -> Union[str, Dict[str, str]]:
    out_dir = work_dir / "out"
    org_dir = work_dir / "org"
    specs_dir = org_dir / "package"
    if specs_dir.exists():
        shutil.rmtree(specs_dir)
    specs_dir.mkdir(parents=True)

    if type(docs) is str:
        (specs_dir / "spec.yaml").write_text(docs)
    else:
        for name, content in docs.items():
            path = specs_dir / name
            path.parent.mkdir(parents=True, exist_ok=True)
            path.write_text(content)

    result = cli_runner(
        make_docs,
        [
            specs_dir.as_posix(),
            out_dir.as_posix(),
            "--spec-root",
            str(work_dir),
        ]
        + (["--flatten" if flatten else "--no-flatten"] if flatten is not None else [])
        + list(args),
    )

    assert result.exit_code == 0, show_cli_trace(result)

    if type(docs) is str:
        return (out_dir / "org" / "package.spec.md").read_text().strip()
    else:
        return {
            file.relative_to(out_dir).as_posix(): file.read_text().strip()
            for file in out_dir.glob("*.md")
        }


def test_bootstrap(cli_runner, work_dir):

    out_yaml = work_dir / "out-spec.yaml"

    result = cli_runner(
        bootstrap,
        [
            str(out_yaml),
            "--author",
            "An Author",
            "$an.author@a.institution",
            "--version",
            "7.1.1",
            "--base-image",
            "name",
            "vnmd/freesurfer_7.1.1",
            "--base-image",
            "tag",
            "20211216",
            "--base-image",
            "package_manager",
            "apt",
            "--packages-pip",
            "fileformats-medimage-extras==0.8.0",
            "--packages-neurodocker",
            "dcm2niix==v1.0.20201102",
            "--command-task",
            "shell::mri_convert <head:medimage/nifti-gz> <brain:medimage/nifti-gz>",
            # "--command-input",
            # "head",
            # "datatype=medimage/nifti-gz,configuration.position=-2,configuration.argstr=''",
            # "--command-output",
            # "brain",
            # "datatype=medimage/nifti-gz,configuration.position=-1,configuration.argstr=''",
            # "--command-configuration",
            # "executable",
            # "mri_convert",
            "--title",
            "MRI Convert",
            "--license",
            "freesurfer",
            "/opt/freesurfer/license.txt",
            "https://surfer.nmr.mgh.harvard.edu/registration.html",
            "The freesurfer license that is required to run any tool within the package",
        ],
    )

    assert result.exit_code == 0, show_cli_trace(result)

    # Check that the YAML loads properly
    App.load(out_yaml)



import typing as ty
from pathlib import Path
from pydra2app.core.image import App
from frametree.core.frameset import FrameSet
from frametree.common import FileSystem, Samples
from pydra2app.testing.constants import (
    ORG,
    LICENSE_CONTENTS,
    LICENSE_NAME,
    LICENSE_INPUT_PATH,
    LICENSE_PATH_PARAM,
)


def get_pipeline_image(license_path: str, app_cls: ty.Type[App] = App) -> App:
    return app_cls(
        title="A test of the license installation",
        name="to_be_overridden",
        org=ORG,
        version="1.0",
        authors=[{"name": "Some One", "email": "some.one@an.email.org"}],
        docs={
            "info_url": "http://concatenate.readthefakedocs.io",
        },
        readme="This is a test README",
        packages={
            "system": ["git", "vim"],
            "pip": ["fileformats", "frametree", "pydra2app", "pydra"],
        },
        licenses={
            LICENSE_NAME: {
                "destination": license_path,
                "info_url": "http://license.test",
                "description": "This is a license to test the build structure",
            }
        },
        commands={
            "check-license": {
                "task": "pydra2app.testing.tasks:CheckLicence",
                "row_frequency": "common:Samples[sample]",
                "parameters": [LICENSE_PATH_PARAM],
                # "inputs": [
                #     {
                #         "name": LICENSE_INPUT_FIELD,
                #         "datatype": "text/plain",
                #         "field": "expected_license_contents",
                #         "help": "the path to the license",
                #     },
                # ],
                # "outputs": [
                #     {
                #         "name": LICENSE_OUTPUT_FIELD,
                #         "datatype": "text/plain",
                #         "field": "out",
                #         "help": "the validated license path",
                #     }
                # ],
                # "parameters": [
                #     {
                #         "name": LICENSE_PATH_PARAM,
                #         "datatype": "field/text",
                #         "field": "expected_license_path",
                #         "required": True,
                #         "help": "the expected contents of the license file",
                #     }
                # ],
            },
        },
    )


def make_dataset(dataset_dir: Path) -> FrameSet:

    contents_dir = dataset_dir / "sample1"
    contents_dir.mkdir(parents=True)

    with open(contents_dir / (LICENSE_INPUT_PATH + ".txt"), "w") as f:
        f.write(LICENSE_CONTENTS)

    dataset = FileSystem().define_frameset(dataset_dir, axes=Samples)
    dataset.save()
    return dataset
